# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import time

import numpy as np
import pytest

try:
    from geometry_msgs.msg import TransformStamped as ROSTransformStamped
except ImportError:
    ROSTransformStamped = None


from dimos.msgs.geometry_msgs import Pose, PoseStamped, Quaternion, Transform, Vector3


def test_transform_initialization() -> None:
    # Test default initialization (identity transform)
    tf = Transform()
    assert tf.translation.x == 0.0
    assert tf.translation.y == 0.0
    assert tf.translation.z == 0.0
    assert tf.rotation.x == 0.0
    assert tf.rotation.y == 0.0
    assert tf.rotation.z == 0.0
    assert tf.rotation.w == 1.0

    # Test initialization with Vector3 and Quaternion
    trans = Vector3(1.0, 2.0, 3.0)
    rot = Quaternion(0.0, 0.0, 0.707107, 0.707107)  # 90 degrees around Z
    tf2 = Transform(translation=trans, rotation=rot)
    assert tf2.translation == trans
    assert tf2.rotation == rot

    # Test initialization with only translation
    tf5 = Transform(translation=Vector3(7.0, 8.0, 9.0))
    assert tf5.translation.x == 7.0
    assert tf5.translation.y == 8.0
    assert tf5.translation.z == 9.0
    assert tf5.rotation.w == 1.0  # Identity rotation

    # Test initialization with only rotation
    tf6 = Transform(rotation=Quaternion(0.0, 0.0, 0.0, 1.0))
    assert tf6.translation.is_zero()  # Zero translation
    assert tf6.rotation.w == 1.0

    # Test keyword argument initialization
    tf7 = Transform(translation=Vector3(1, 2, 3), rotation=Quaternion())
    assert tf7.translation == Vector3(1, 2, 3)
    assert tf7.rotation == Quaternion()

    # Test keyword with only translation
    tf8 = Transform(translation=Vector3(4, 5, 6))
    assert tf8.translation == Vector3(4, 5, 6)
    assert tf8.rotation.w == 1.0

    # Test keyword with only rotation
    tf9 = Transform(rotation=Quaternion(0, 0, 1, 0))
    assert tf9.translation.is_zero()
    assert tf9.rotation == Quaternion(0, 0, 1, 0)


def test_transform_identity() -> None:
    # Test identity class method
    tf = Transform.identity()
    assert tf.translation.is_zero()
    assert tf.rotation.x == 0.0
    assert tf.rotation.y == 0.0
    assert tf.rotation.z == 0.0
    assert tf.rotation.w == 1.0

    # Identity should equal default constructor
    assert tf == Transform()


def test_transform_equality() -> None:
    tf1 = Transform(translation=Vector3(1, 2, 3), rotation=Quaternion(0, 0, 0, 1))
    tf2 = Transform(translation=Vector3(1, 2, 3), rotation=Quaternion(0, 0, 0, 1))
    tf3 = Transform(translation=Vector3(1, 2, 4), rotation=Quaternion(0, 0, 0, 1))  # Different z
    tf4 = Transform(
        translation=Vector3(1, 2, 3), rotation=Quaternion(0, 0, 1, 0)
    )  # Different rotation

    assert tf1 == tf2
    assert tf1 != tf3
    assert tf1 != tf4
    assert tf1 != "not a transform"


def test_transform_string_representations() -> None:
    tf = Transform(
        translation=Vector3(1.5, -2.0, 3.14), rotation=Quaternion(0, 0, 0.707107, 0.707107)
    )

    # Test repr
    repr_str = repr(tf)
    assert "Transform" in repr_str
    assert "translation=" in repr_str
    assert "rotation=" in repr_str
    assert "1.5" in repr_str

    # Test str
    str_str = str(tf)
    assert "Translation:" in str_str
    assert "Rotation:" in str_str


def test_pose_add_transform() -> None:
    initial_pose = Pose(1.0, 0.0, 0.0)

    # 90 degree rotation around Z axis
    angle = np.pi / 2
    transform = Transform(
        translation=Vector3(2.0, 1.0, 0.0),
        rotation=Quaternion(0.0, 0.0, np.sin(angle / 2), np.cos(angle / 2)),
    )

    transformed_pose = initial_pose @ transform

    # - Translation (2, 1, 0) is added directly to position (1, 0, 0)
    # - Result position: (3, 1, 0)
    assert np.isclose(transformed_pose.position.x, 3.0, atol=1e-10)
    assert np.isclose(transformed_pose.position.y, 1.0, atol=1e-10)
    assert np.isclose(transformed_pose.position.z, 0.0, atol=1e-10)

    # Rotation should be 90 degrees around Z
    assert np.isclose(transformed_pose.orientation.x, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose.orientation.y, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose.orientation.z, np.sin(angle / 2), atol=1e-10)
    assert np.isclose(transformed_pose.orientation.w, np.cos(angle / 2), atol=1e-10)

    initial_pose_stamped = PoseStamped(
        position=initial_pose.position, orientation=initial_pose.orientation
    )
    transformed_pose_stamped = PoseStamped(
        position=transformed_pose.position, orientation=transformed_pose.orientation
    )

    found_tf = initial_pose_stamped.find_transform(transformed_pose_stamped)

    assert found_tf.translation == transform.translation
    assert found_tf.rotation == transform.rotation
    assert found_tf.translation.x == transform.translation.x
    assert found_tf.translation.y == transform.translation.y
    assert found_tf.translation.z == transform.translation.z

    assert found_tf.rotation.x == transform.rotation.x
    assert found_tf.rotation.y == transform.rotation.y
    assert found_tf.rotation.z == transform.rotation.z
    assert found_tf.rotation.w == transform.rotation.w

    print(found_tf.rotation, found_tf.translation)


def test_pose_add_transform_with_rotation() -> None:
    # Create a pose at (0, 0, 0) rotated 90 degrees around Z
    angle = np.pi / 2
    initial_pose = Pose(0.0, 0.0, 0.0, 0.0, 0.0, np.sin(angle / 2), np.cos(angle / 2))

    # Add 45 degree rotation to transform1
    rotation_angle = np.pi / 4  # 45 degrees
    transform1 = Transform(
        translation=Vector3(1.0, 0.0, 0.0),
        rotation=Quaternion(
            0.0, 0.0, np.sin(rotation_angle / 2), np.cos(rotation_angle / 2)
        ),  # 45� around Z
    )

    transform2 = Transform(
        translation=Vector3(0.0, 1.0, 1.0),
        rotation=Quaternion(0.0, 0.0, 0.0, 1.0),  # No rotation
    )

    transformed_pose1 = initial_pose @ transform1
    transformed_pose2 = initial_pose @ transform1 @ transform2

    # Test transformed_pose1: initial_pose + transform1
    # Since the pose is rotated 90� (facing +Y), moving forward (local X)
    # means moving in the +Y direction in world frame
    assert np.isclose(transformed_pose1.position.x, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose1.position.y, 1.0, atol=1e-10)
    assert np.isclose(transformed_pose1.position.z, 0.0, atol=1e-10)

    # Orientation should be 90� + 45� = 135� around Z
    total_angle1 = angle + rotation_angle  # 135 degrees
    assert np.isclose(transformed_pose1.orientation.x, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose1.orientation.y, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose1.orientation.z, np.sin(total_angle1 / 2), atol=1e-10)
    assert np.isclose(transformed_pose1.orientation.w, np.cos(total_angle1 / 2), atol=1e-10)

    # Test transformed_pose2: initial_pose + transform1 + transform2
    # Starting from (0, 0, 0) facing 90�:
    #
    # - Apply transform1: move 1 forward (along +Y) � (0, 1, 0), now facing 135�
    #
    # - Apply transform2: move 1 in local Y and 1 up
    #   At 135�, local Y points at 225� (135� + 90�)
    #
    #   x += cos(225�) = -2/2, y += sin(225�) = -2/2
    sqrt2_2 = np.sqrt(2) / 2
    expected_x = 0.0 - sqrt2_2  # 0 - 2/2 H -0.707
    expected_y = 1.0 - sqrt2_2  # 1 - 2/2 H 0.293
    expected_z = 1.0  # 0 + 1

    assert np.isclose(transformed_pose2.position.x, expected_x, atol=1e-10)
    assert np.isclose(transformed_pose2.position.y, expected_y, atol=1e-10)
    assert np.isclose(transformed_pose2.position.z, expected_z, atol=1e-10)

    # Orientation should be 135� (only transform1 has rotation)
    total_angle2 = total_angle1  # 135 degrees (transform2 has no rotation)
    assert np.isclose(transformed_pose2.orientation.x, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose2.orientation.y, 0.0, atol=1e-10)
    assert np.isclose(transformed_pose2.orientation.z, np.sin(total_angle2 / 2), atol=1e-10)
    assert np.isclose(transformed_pose2.orientation.w, np.cos(total_angle2 / 2), atol=1e-10)


def test_lcm_encode_decode() -> None:
    angle = np.pi / 2
    transform = Transform(
        translation=Vector3(2.0, 1.0, 0.0),
        rotation=Quaternion(0.0, 0.0, np.sin(angle / 2), np.cos(angle / 2)),
    )

    data = transform.lcm_encode()

    decoded_transform = Transform.lcm_decode(data)

    assert decoded_transform == transform


def test_transform_addition() -> None:
    # Test 1: Simple translation addition (no rotation)
    t1 = Transform(
        translation=Vector3(1, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),  # identity rotation
    )
    t2 = Transform(
        translation=Vector3(2, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),  # identity rotation
    )
    t3 = t1 + t2
    assert t3.translation == Vector3(3, 0, 0)
    assert t3.rotation == Quaternion(0, 0, 0, 1)

    # Test 2: 90-degree rotation composition
    # First transform: move 1 unit in X
    t1 = Transform(
        translation=Vector3(1, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),  # identity
    )
    # Second transform: move 1 unit in X with 90-degree rotation around Z
    angle = np.pi / 2
    t2 = Transform(
        translation=Vector3(1, 0, 0),
        rotation=Quaternion(0, 0, np.sin(angle / 2), np.cos(angle / 2)),
    )
    t3 = t1 + t2
    assert t3.translation == Vector3(2, 0, 0)
    # Rotation should be 90 degrees around Z
    assert np.isclose(t3.rotation.x, 0.0, atol=1e-10)
    assert np.isclose(t3.rotation.y, 0.0, atol=1e-10)
    assert np.isclose(t3.rotation.z, np.sin(angle / 2), atol=1e-10)
    assert np.isclose(t3.rotation.w, np.cos(angle / 2), atol=1e-10)

    # Test 3: Rotation affects translation
    # First transform: 90-degree rotation around Z
    t1 = Transform(
        translation=Vector3(0, 0, 0),
        rotation=Quaternion(0, 0, np.sin(angle / 2), np.cos(angle / 2)),  # 90° around Z
    )
    # Second transform: move 1 unit in X
    t2 = Transform(
        translation=Vector3(1, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),  # identity
    )
    t3 = t1 + t2
    # X direction rotated 90° becomes Y direction
    assert np.isclose(t3.translation.x, 0.0, atol=1e-10)
    assert np.isclose(t3.translation.y, 1.0, atol=1e-10)
    assert np.isclose(t3.translation.z, 0.0, atol=1e-10)
    # Rotation remains 90° around Z
    assert np.isclose(t3.rotation.z, np.sin(angle / 2), atol=1e-10)
    assert np.isclose(t3.rotation.w, np.cos(angle / 2), atol=1e-10)

    # Test 4: Frame tracking
    t1 = Transform(
        translation=Vector3(1, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),
        frame_id="world",
        child_frame_id="robot",
    )
    t2 = Transform(
        translation=Vector3(2, 0, 0),
        rotation=Quaternion(0, 0, 0, 1),
        frame_id="robot",
        child_frame_id="sensor",
    )
    t3 = t1 + t2
    assert t3.frame_id == "world"
    assert t3.child_frame_id == "sensor"

    # Test 5: Type error
    with pytest.raises(TypeError):
        t1 + "not a transform"


def test_transform_from_pose() -> None:
    """Test converting Pose to Transform"""
    # Create a Pose with position and orientation
    pose = Pose(
        position=Vector3(1.0, 2.0, 3.0),
        orientation=Quaternion(0.0, 0.0, 0.707, 0.707),  # 90 degrees around Z
    )

    # Convert to Transform
    transform = Transform.from_pose("base_link", pose)

    # Check that translation and rotation match
    assert transform.translation == pose.position
    assert transform.rotation == pose.orientation
    assert transform.frame_id == "world"  # default frame_id
    assert transform.child_frame_id == "base_link"  # passed as first argument


# validating results from example @
# https://foxglove.dev/blog/understanding-ros-transforms
def test_transform_from_ros() -> None:
    """Test converting PoseStamped to Transform"""
    test_time = time.time()
    pose_stamped = PoseStamped(
        ts=test_time,
        frame_id="base_link",
        position=Vector3(1, -1, 0),
        orientation=Quaternion.from_euler(Vector3(0, 0, math.pi / 6)),
    )
    transform_base_link_to_arm = Transform.from_pose("arm_base_link", pose_stamped)

    transform_arm_to_end = Transform.from_pose(
        "end",
        PoseStamped(
            ts=test_time,
            frame_id="arm_base_link",
            position=Vector3(1, 1, 0),
            orientation=Quaternion.from_euler(Vector3(0, 0, math.pi / 6)),
        ),
    )

    print(transform_base_link_to_arm)
    print(transform_arm_to_end)

    end_effector_global_pose = transform_base_link_to_arm + transform_arm_to_end

    assert end_effector_global_pose.translation.x == pytest.approx(1.366, abs=1e-3)
    assert end_effector_global_pose.translation.y == pytest.approx(0.366, abs=1e-3)


def test_transform_from_pose_stamped() -> None:
    """Test converting PoseStamped to Transform"""
    # Create a PoseStamped with position, orientation, timestamp and frame
    test_time = time.time()
    pose_stamped = PoseStamped(
        ts=test_time,
        frame_id="map",
        position=Vector3(4.0, 5.0, 6.0),
        orientation=Quaternion(0.0, 0.707, 0.0, 0.707),  # 90 degrees around Y
    )

    # Convert to Transform
    transform = Transform.from_pose("robot_base", pose_stamped)

    # Check that all fields match
    assert transform.translation == pose_stamped.position
    assert transform.rotation == pose_stamped.orientation
    assert transform.frame_id == pose_stamped.frame_id
    assert transform.ts == pose_stamped.ts
    assert transform.child_frame_id == "robot_base"  # passed as first argument


def test_transform_from_pose_variants() -> None:
    """Test from_pose with different Pose initialization methods"""
    # Test with Pose created from x,y,z
    pose1 = Pose(1.0, 2.0, 3.0)
    transform1 = Transform.from_pose("base_link", pose1)
    assert transform1.translation.x == 1.0
    assert transform1.translation.y == 2.0
    assert transform1.translation.z == 3.0
    assert transform1.rotation.w == 1.0  # Identity quaternion

    # Test with Pose created from tuple
    pose2 = Pose(([7.0, 8.0, 9.0], [0.0, 0.0, 0.0, 1.0]))
    transform2 = Transform.from_pose("base_link", pose2)
    assert transform2.translation.x == 7.0
    assert transform2.translation.y == 8.0
    assert transform2.translation.z == 9.0

    # Test with Pose created from dict
    pose3 = Pose({"position": [10.0, 11.0, 12.0], "orientation": [0.0, 0.0, 0.0, 1.0]})
    transform3 = Transform.from_pose("base_link", pose3)
    assert transform3.translation.x == 10.0
    assert transform3.translation.y == 11.0
    assert transform3.translation.z == 12.0


def test_transform_from_pose_invalid_type() -> None:
    """Test that from_pose raises TypeError for invalid types"""
    with pytest.raises(TypeError):
        Transform.from_pose("not a pose")

    with pytest.raises(TypeError):
        Transform.from_pose(42)

    with pytest.raises(TypeError):
        Transform.from_pose(None)


@pytest.mark.ros
def test_transform_from_ros_transform_stamped() -> None:
    """Test creating a Transform from a ROS TransformStamped message."""
    ros_msg = ROSTransformStamped()
    ros_msg.header.frame_id = "world"
    ros_msg.header.stamp.sec = 123
    ros_msg.header.stamp.nanosec = 456000000
    ros_msg.child_frame_id = "robot"
    ros_msg.transform.translation.x = 1.0
    ros_msg.transform.translation.y = 2.0
    ros_msg.transform.translation.z = 3.0
    ros_msg.transform.rotation.x = 0.1
    ros_msg.transform.rotation.y = 0.2
    ros_msg.transform.rotation.z = 0.3
    ros_msg.transform.rotation.w = 0.9

    transform = Transform.from_ros_transform_stamped(ros_msg)

    assert transform.frame_id == "world"
    assert transform.child_frame_id == "robot"
    assert transform.ts == 123.456
    assert transform.translation.x == 1.0
    assert transform.translation.y == 2.0
    assert transform.translation.z == 3.0
    assert transform.rotation.x == 0.1
    assert transform.rotation.y == 0.2
    assert transform.rotation.z == 0.3
    assert transform.rotation.w == 0.9


@pytest.mark.ros
def test_transform_to_ros_transform_stamped() -> None:
    """Test converting a Transform to a ROS TransformStamped message."""
    transform = Transform(
        translation=Vector3(4.0, 5.0, 6.0),
        rotation=Quaternion(0.15, 0.25, 0.35, 0.85),
        frame_id="base_link",
        child_frame_id="sensor",
        ts=124.789,
    )

    ros_msg = transform.to_ros_transform_stamped()

    assert isinstance(ros_msg, ROSTransformStamped)
    assert ros_msg.header.frame_id == "base_link"
    assert ros_msg.child_frame_id == "sensor"
    assert ros_msg.header.stamp.sec == 124
    assert ros_msg.header.stamp.nanosec == 789000000
    assert ros_msg.transform.translation.x == 4.0
    assert ros_msg.transform.translation.y == 5.0
    assert ros_msg.transform.translation.z == 6.0
    assert ros_msg.transform.rotation.x == 0.15
    assert ros_msg.transform.rotation.y == 0.25
    assert ros_msg.transform.rotation.z == 0.35
    assert ros_msg.transform.rotation.w == 0.85


@pytest.mark.ros
def test_transform_ros_roundtrip() -> None:
    """Test round-trip conversion between Transform and ROS TransformStamped."""
    original = Transform(
        translation=Vector3(7.5, 8.5, 9.5),
        rotation=Quaternion(0.0, 0.0, 0.383, 0.924),  # ~45 degrees around Z
        frame_id="odom",
        child_frame_id="base_footprint",
        ts=99.123,
    )

    ros_msg = original.to_ros_transform_stamped()
    restored = Transform.from_ros_transform_stamped(ros_msg)

    assert restored.frame_id == original.frame_id
    assert restored.child_frame_id == original.child_frame_id
    assert restored.ts == original.ts
    assert restored.translation.x == original.translation.x
    assert restored.translation.y == original.translation.y
    assert restored.translation.z == original.translation.z
    assert restored.rotation.x == original.rotation.x
    assert restored.rotation.y == original.rotation.y
    assert restored.rotation.z == original.rotation.z
    assert restored.rotation.w == original.rotation.w



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""URDF generation utilities."""

from __future__ import annotations


def box_urdf(
    width: float,
    height: float,
    depth: float,
    name: str = "box_robot",
    mass: float = 1.0,
    rgba: tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.5),
) -> str:
    """Generate a simple URDF with a box as the base_link.

    Args:
        width: Box size in X direction (meters)
        height: Box size in Y direction (meters)
        depth: Box size in Z direction (meters)
        name: Robot name
        mass: Mass of the box (kg)
        rgba: Color as (red, green, blue, alpha), default red with 0.5 transparency

    Returns:
        URDF XML string
    """
    # Simple box inertia (solid cuboid)
    ixx = (mass / 12.0) * (height**2 + depth**2)
    iyy = (mass / 12.0) * (width**2 + depth**2)
    izz = (mass / 12.0) * (width**2 + height**2)

    r, g, b, a = rgba
    return f"""<?xml version="1.0"?>
<robot name="{name}">
  <link name="base_link">
    <visual>
      <geometry>
        <box size="{width} {height} {depth}"/>
      </geometry>
      <material name="box_material">
        <color rgba="{r} {g} {b} {a}"/>
      </material>
    </visual>
    <collision>
      <geometry>
        <box size="{width} {height} {depth}"/>
      </geometry>
    </collision>
    <inertial>
      <mass value="{mass}"/>
      <inertia ixx="{ixx:.6f}" ixy="0" ixz="0" iyy="{iyy:.6f}" iyz="0" izz="{izz:.6f}"/>
    </inertial>
  </link>
</robot>
"""



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joint Trajectory Controller Specification

A simple joint-space trajectory executor. Does NOT:
- Use Cartesian space
- Compute error
- Apply PID
- Call IK

Just samples a trajectory at time t and sends joint positions to the driver.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from dimos.core import In, Out
    from dimos.msgs.sensor_msgs import JointCommand, JointState, RobotState
    from dimos.msgs.trajectory_msgs import JointTrajectory as JointTrajectoryMsg, TrajectoryState

# Input topics
joint_state: In[JointState] | None = None  # Feedback from arm driver
robot_state: In[RobotState] | None = None  # Robot status from arm driver
trajectory: In[JointTrajectoryMsg] | None = None  # Desired trajectory

# Output topics
joint_position_command: Out[JointCommand] | None = None  # To arm driver


def execute_trajectory() -> bool:
    """
    Set and start executing a new trajectory immediately.
    Returns True if accepted, False if controller busy or traj invalid.
    """
    raise NotImplementedError("Protocol method")


def cancel() -> bool:
    """
    Cancel the currently executing trajectory.
    Returns True if cancelled, False if no active trajectory.
    """
    raise NotImplementedError("Protocol method")


def get_status() -> TrajectoryStatusProtocol:
    """
    Get the current status of the trajectory execution.
    Returns a TrajectoryStatus message with details.
     "state": "IDLE" | "EXECUTING" | "COMPLETED" | "ABORTED" | "FAULT",
      "progress": float in [0,1],
      "active_traj_id": Optional[str],
      "error": Optional[str],
    """
    raise NotImplementedError("Protocol method")
    ...


class JointTrajectoryProtocol(Protocol):
    """Protocol for a joint trajectory object."""

    duration: float  # Total duration in seconds

    def sample(self, t: float) -> tuple[list[float], list[float]]:
        """
        Sample the trajectory at time t.

        Args:
            t: Time in seconds (0 <= t <= duration)

        Returns:
            Tuple of (q_ref, qd_ref):
                - q_ref: Joint positions (radians)
                - qd_ref: Joint velocities (rad/s)
        """
        ...


class TrajectoryStatusProtocol(Protocol):
    """Status of trajectory execution."""

    state: TrajectoryState  # Current state
    progress: float  # Progress 0.0 to 1.0
    time_elapsed: float  # Seconds since trajectory start
    time_remaining: float  # Estimated seconds remaining
    error: str | None  # Error message if FAULT state



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import datetime
import json
from operator import itemgetter
import os
from typing import Any, TypedDict
import uuid

from langchain_core.messages import (
    AIMessage,
    HumanMessage,
    SystemMessage,
    ToolCall,
    ToolMessage,
)

from dimos.agents.llm_init import build_llm, build_system_message
from dimos.agents.spec import AgentSpec, Model, Provider
from dimos.core import DimosCluster, rpc
from dimos.protocol.skill.coordinator import SkillCoordinator, SkillState, SkillStateDict
from dimos.protocol.skill.skill import SkillContainer
from dimos.protocol.skill.type import Output
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


SYSTEM_MSG_APPEND = "\nYour message history will always be appended with a System Overview message that provides situational awareness."


def toolmsg_from_state(state: SkillState) -> ToolMessage:
    if state.skill_config.output != Output.standard:
        content = "output attached in separate messages"
    else:
        content = state.content()  # type: ignore[assignment]

    return ToolMessage(
        # if agent call has been triggered by another skill,
        # and this specific skill didn't finish yet but we need a tool call response
        # we return a message explaining that execution is still ongoing
        content=content
        or "Running, you will be called with an update, no need for subsequent tool calls",
        name=state.name,
        tool_call_id=state.call_id,
    )


class SkillStateSummary(TypedDict):
    name: str
    call_id: str
    state: str
    data: Any


def summary_from_state(state: SkillState, special_data: bool = False) -> SkillStateSummary:
    content = state.content()
    if isinstance(content, dict):
        content = json.dumps(content)

    if not isinstance(content, str):
        content = str(content)

    return {
        "name": state.name,
        "call_id": state.call_id,
        "state": state.state.name,
        "data": state.content() if not special_data else "data will be in a separate message",
    }


def _custom_json_serializers(obj):  # type: ignore[no-untyped-def]
    if isinstance(obj, datetime.date | datetime.datetime):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


# takes an overview of running skills from the coorindator
# and builds messages to be sent to an agent
def snapshot_to_messages(
    state: SkillStateDict,
    tool_calls: list[ToolCall],
) -> tuple[list[ToolMessage], AIMessage | None]:
    # builds a set of tool call ids from a previous agent request
    tool_call_ids = set(
        map(itemgetter("id"), tool_calls),
    )

    # build a tool msg responses
    tool_msgs: list[ToolMessage] = []

    # build a general skill state overview (for longer running skills)
    state_overview: list[dict[str, SkillStateSummary]] = []

    # for special skills that want to return a separate message
    # (images for example, requires to be a HumanMessage)
    special_msgs: list[HumanMessage] = []

    # for special skills that want to return a separate message that should
    # stay in history, like actual human messages, critical events
    history_msgs: list[HumanMessage] = []

    # Initialize state_msg
    state_msg = None

    for skill_state in sorted(
        state.values(),
        key=lambda skill_state: skill_state.duration(),
    ):
        if skill_state.call_id in tool_call_ids:
            tool_msgs.append(toolmsg_from_state(skill_state))

        if skill_state.skill_config.output == Output.human:
            content = skill_state.content()
            if not content:
                continue
            history_msgs.append(HumanMessage(content=content))  # type: ignore[arg-type]
            continue

        special_data = skill_state.skill_config.output == Output.image
        if special_data:
            content = skill_state.content()
            if not content:
                continue
            special_msgs.append(HumanMessage(content=content))  # type: ignore[arg-type]

        if skill_state.call_id in tool_call_ids:
            continue

        state_overview.append(summary_from_state(skill_state, special_data))  # type: ignore[arg-type]

    if state_overview:
        state_overview_str = "\n".join(
            json.dumps(s, default=_custom_json_serializers) for s in state_overview
        )
        state_msg = AIMessage("State Overview:\n" + state_overview_str)

    return {  # type: ignore[return-value]
        "tool_msgs": tool_msgs,
        "history_msgs": history_msgs,
        "state_msgs": ([state_msg] if state_msg else []) + special_msgs,
    }


# Agent class job is to glue skill coordinator state to an agent, builds langchain messages
class Agent(AgentSpec):
    system_message: SystemMessage
    state_messages: list[AIMessage | HumanMessage]

    def __init__(  # type: ignore[no-untyped-def]
        self,
        *args,
        **kwargs,
    ) -> None:
        AgentSpec.__init__(self, *args, **kwargs)

        self.state_messages = []
        self.coordinator = SkillCoordinator()
        self._history = []  # type: ignore[var-annotated]
        self._agent_id = str(uuid.uuid4())
        self._agent_stopped = False

        self.system_message = build_system_message(self.config, append=SYSTEM_MSG_APPEND)
        self.publish(self.system_message)
        self._llm = build_llm(self.config)

    @rpc
    def get_agent_id(self) -> str:
        return self._agent_id

    @rpc
    def start(self) -> None:
        super().start()
        self.coordinator.start()

    @rpc
    def stop(self) -> None:
        self.coordinator.stop()
        self._agent_stopped = True
        super().stop()

    def clear_history(self) -> None:
        self._history.clear()

    def append_history(self, *msgs: list[AIMessage | HumanMessage]) -> None:
        for msg in msgs:
            self.publish(msg)  # type: ignore[arg-type]

        self._history.extend(msgs)

    def history(self):  # type: ignore[no-untyped-def]
        return [self.system_message, *self._history, *self.state_messages]

    # Used by agent to execute tool calls
    def execute_tool_calls(self, tool_calls: list[ToolCall]) -> None:
        """Execute a list of tool calls from the agent."""
        if self._agent_stopped:
            logger.warning("Agent is stopped, cannot execute tool calls.")
            return
        for tool_call in tool_calls:
            logger.info(f"executing skill call {tool_call}")
            self.coordinator.call_skill(
                tool_call.get("id"),  # type: ignore[arg-type]
                tool_call.get("name"),
                tool_call.get("args"),
            )

    # used to inject skill calls into the agent loop without agent asking for it
    def run_implicit_skill(self, skill_name: str, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if self._agent_stopped:
            logger.warning("Agent is stopped, cannot execute implicit skill calls.")
            return
        self.coordinator.call_skill(False, skill_name, {"args": kwargs})

    async def agent_loop(self, first_query: str = ""):  # type: ignore[no-untyped-def]
        # TODO: Should I add a lock here to prevent concurrent calls to agent_loop?

        if self._agent_stopped:
            logger.warning("Agent is stopped, cannot run agent loop.")
            # return "Agent is stopped."
            import traceback

            traceback.print_stack()
            return "Agent is stopped."

        self.state_messages = []
        if first_query:
            self.append_history(HumanMessage(first_query))  # type: ignore[arg-type]

        def _get_state() -> str:
            # TODO: FIX THIS EXTREME HACK
            update = self.coordinator.generate_snapshot(clear=False)
            snapshot_msgs = snapshot_to_messages(update, msg.tool_calls)
            return json.dumps(snapshot_msgs, sort_keys=True, default=lambda o: repr(o))

        try:
            while True:
                # we are getting tools from the coordinator on each turn
                # since this allows for skillcontainers to dynamically provide new skills
                tools = self.get_tools()  # type: ignore[no-untyped-call]
                self._llm = self._llm.bind_tools(tools)  # type: ignore[assignment]

                # publish to /agent topic for observability
                for state_msg in self.state_messages:
                    self.publish(state_msg)

                # history() builds our message history dynamically
                # ensures we include latest system state, but not old ones.
                messages = self.history()  # type: ignore[no-untyped-call]

                # Some LLMs don't work without any human messages. Add an initial one.
                if len(messages) == 1 and isinstance(messages[0], SystemMessage):
                    messages.append(
                        HumanMessage(
                            "Everything is initialized. I'll let you know when you should act."
                        )
                    )
                    self.append_history(messages[-1])

                msg = self._llm.invoke(messages)

                self.append_history(msg)  # type: ignore[arg-type]

                logger.info(f"Agent response: {msg.content}")

                state = _get_state()

                if msg.tool_calls:
                    self.execute_tool_calls(msg.tool_calls)

                # print(self)
                # print(self.coordinator)

                self._write_debug_history_file()

                if not self.coordinator.has_active_skills():
                    logger.info("No active tasks, exiting agent loop.")
                    return msg.content

                # coordinator will continue once a skill state has changed in
                # such a way that agent call needs to be executed

                if state == _get_state():
                    await self.coordinator.wait_for_updates()

                # we request a full snapshot of currently running, finished or errored out skills
                # we ask for removal of finished skills from subsequent snapshots (clear=True)
                update = self.coordinator.generate_snapshot(clear=True)

                # generate tool_msgs and general state update message,
                # depending on a skill having associated tool call from previous interaction
                # we will return a tool message, and not a general state message
                snapshot_msgs = snapshot_to_messages(update, msg.tool_calls)

                self.state_messages = snapshot_msgs.get("state_msgs", [])  # type: ignore[attr-defined]
                self.append_history(
                    *snapshot_msgs.get("tool_msgs", []),  # type: ignore[attr-defined]
                    *snapshot_msgs.get("history_msgs", []),  # type: ignore[attr-defined]
                )

        except Exception as e:
            logger.error(f"Error in agent loop: {e}")
            import traceback

            traceback.print_exc()

    @rpc
    def loop_thread(self) -> bool:
        asyncio.run_coroutine_threadsafe(self.agent_loop(), self._loop)  # type: ignore[arg-type]
        return True

    @rpc
    def query(self, query: str):  # type: ignore[no-untyped-def]
        # TODO: could this be
        # from distributed.utils import sync
        # return sync(self._loop, self.agent_loop, query)
        return asyncio.run_coroutine_threadsafe(self.agent_loop(query), self._loop).result()  # type: ignore[arg-type]

    async def query_async(self, query: str):  # type: ignore[no-untyped-def]
        return await self.agent_loop(query)

    @rpc
    def register_skills(self, container, run_implicit_name: str | None = None):  # type: ignore[no-untyped-def]
        ret = self.coordinator.register_skills(container)  # type: ignore[func-returns-value]

        if run_implicit_name:
            self.run_implicit_skill(run_implicit_name)

        return ret

    def get_tools(self):  # type: ignore[no-untyped-def]
        return self.coordinator.get_tools()

    def _write_debug_history_file(self) -> None:
        file_path = os.getenv("DEBUG_AGENT_HISTORY_FILE")
        if not file_path:
            return

        history = [x.__dict__ for x in self.history()]  # type: ignore[no-untyped-call]

        with open(file_path, "w") as f:
            json.dump(history, f, default=lambda x: repr(x), indent=2)


class LlmAgent(Agent):
    @rpc
    def start(self) -> None:
        super().start()
        self.loop_thread()

    @rpc
    def stop(self) -> None:
        super().stop()


llm_agent = LlmAgent.blueprint


def deploy(
    dimos: DimosCluster,
    system_prompt: str = "You are a helpful assistant for controlling a Unitree Go2 robot.",
    model: Model = Model.GPT_4O,
    provider: Provider = Provider.OPENAI,  # type: ignore[attr-defined]
    skill_containers: list[SkillContainer] | None = None,
) -> Agent:
    from dimos.agents.cli.human import HumanInput

    if skill_containers is None:
        skill_containers = []
    agent = dimos.deploy(  # type: ignore[attr-defined]
        Agent,
        system_prompt=system_prompt,
        model=model,
        provider=provider,
    )

    human_input = dimos.deploy(HumanInput)  # type: ignore[attr-defined]
    human_input.start()

    agent.register_skills(human_input)

    for skill_container in skill_containers:
        print("Registering skill container:", skill_container)
        agent.register_skills(skill_container)

    agent.run_implicit_skill("human")
    agent.start()
    agent.loop_thread()

    return agent  # type: ignore[no-any-return]


__all__ = ["Agent", "deploy", "llm_agent"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
from collections.abc import Sequence
from typing import TypeVar, Union

import numpy as np

from dimos.types.ros_polyfill import Vector3

T = TypeVar("T", bound="Vector")

# Vector-like types that can be converted to/from Vector
VectorLike = Union[Sequence[int | float], Vector3, "Vector", np.ndarray]  # type: ignore[type-arg]


class Vector:
    """A wrapper around numpy arrays for vector operations with intuitive syntax."""

    def __init__(self, *args: VectorLike) -> None:
        """Initialize a vector from components or another iterable.

        Examples:
            Vector(1, 2)       # 2D vector
            Vector(1, 2, 3)    # 3D vector
            Vector([1, 2, 3])  # From list
            Vector(np.array([1, 2, 3])) # From numpy array
        """
        if len(args) == 1 and hasattr(args[0], "__iter__"):
            self._data = np.array(args[0], dtype=float)

        elif len(args) == 1:
            self._data = np.array([args[0].x, args[0].y, args[0].z], dtype=float)  # type: ignore[union-attr]

        else:
            self._data = np.array(args, dtype=float)

    @property
    def yaw(self) -> float:
        return self.x

    @property
    def tuple(self) -> tuple[float, ...]:
        """Tuple representation of the vector."""
        return tuple(self._data)

    @property
    def x(self) -> float:
        """X component of the vector."""
        return self._data[0] if len(self._data) > 0 else 0.0

    @property
    def y(self) -> float:
        """Y component of the vector."""
        return self._data[1] if len(self._data) > 1 else 0.0

    @property
    def z(self) -> float:
        """Z component of the vector."""
        return self._data[2] if len(self._data) > 2 else 0.0

    @property
    def dim(self) -> int:
        """Dimensionality of the vector."""
        return len(self._data)

    @property
    def data(self) -> np.ndarray:  # type: ignore[type-arg]
        """Get the underlying numpy array."""
        return self._data

    def __getitem__(self, idx: int):  # type: ignore[no-untyped-def]
        return self._data[idx]

    def __repr__(self) -> str:
        return f"Vector({self.data})"

    def __str__(self) -> str:
        if self.dim < 2:
            return self.__repr__()

        def getArrow():  # type: ignore[no-untyped-def]
            repr = ["←", "↖", "↑", "↗", "→", "↘", "↓", "↙"]

            if self.x == 0 and self.y == 0:
                return "·"

            # Calculate angle in radians and convert to directional index
            angle = np.arctan2(self.y, self.x)
            # Map angle to 0-7 index (8 directions) with proper orientation
            dir_index = int(((angle + np.pi) * 4 / np.pi) % 8)
            # Get directional arrow symbol
            return repr[dir_index]

        return f"{getArrow()} Vector {self.__repr__()}"  # type: ignore[no-untyped-call]

    def serialize(self) -> builtins.tuple:  # type: ignore[type-arg]
        """Serialize the vector to a tuple."""
        return {"type": "vector", "c": self._data.tolist()}  # type: ignore[return-value]

    def __eq__(self, other) -> bool:  # type: ignore[no-untyped-def]
        """Check if two vectors are equal using numpy's allclose for floating point comparison."""
        if not isinstance(other, Vector):
            return False
        if len(self._data) != len(other._data):
            return False
        return np.allclose(self._data, other._data)

    def __add__(self: T, other: VectorLike) -> T:
        other = to_vector(other)
        if self.dim != other.dim:
            max_dim = max(self.dim, other.dim)
            return self.pad(max_dim) + other.pad(max_dim)
        return self.__class__(self._data + other._data)

    def __sub__(self: T, other: VectorLike) -> T:
        other = to_vector(other)
        if self.dim != other.dim:
            max_dim = max(self.dim, other.dim)
            return self.pad(max_dim) - other.pad(max_dim)
        return self.__class__(self._data - other._data)

    def __mul__(self: T, scalar: float) -> T:
        return self.__class__(self._data * scalar)

    def __rmul__(self: T, scalar: float) -> T:
        return self.__mul__(scalar)

    def __truediv__(self: T, scalar: float) -> T:
        return self.__class__(self._data / scalar)

    def __neg__(self: T) -> T:
        return self.__class__(-self._data)

    def dot(self, other: VectorLike) -> float:
        """Compute dot product."""
        other = to_vector(other)
        return float(np.dot(self._data, other._data))

    def cross(self: T, other: VectorLike) -> T:
        """Compute cross product (3D vectors only)."""
        if self.dim != 3:
            raise ValueError("Cross product is only defined for 3D vectors")

        other = to_vector(other)
        if other.dim != 3:
            raise ValueError("Cross product requires two 3D vectors")

        return self.__class__(np.cross(self._data, other._data))

    def length(self) -> float:
        """Compute the Euclidean length (magnitude) of the vector."""
        return float(np.linalg.norm(self._data))

    def length_squared(self) -> float:
        """Compute the squared length of the vector (faster than length())."""
        return float(np.sum(self._data * self._data))

    def normalize(self: T) -> T:
        """Return a normalized unit vector in the same direction."""
        length = self.length()
        if length < 1e-10:  # Avoid division by near-zero
            return self.__class__(np.zeros_like(self._data))
        return self.__class__(self._data / length)

    def to_2d(self: T) -> T:
        """Convert a vector to a 2D vector by taking only the x and y components."""
        return self.__class__(self._data[:2])

    def pad(self: T, dim: int) -> T:
        """Pad a vector with zeros to reach the specified dimension.

        If vector already has dimension >= dim, it is returned unchanged.
        """
        if self.dim >= dim:
            return self

        padded = np.zeros(dim, dtype=float)
        padded[: len(self._data)] = self._data
        return self.__class__(padded)

    def distance(self, other: VectorLike) -> float:
        """Compute Euclidean distance to another vector."""
        other = to_vector(other)
        return float(np.linalg.norm(self._data - other._data))

    def distance_squared(self, other: VectorLike) -> float:
        """Compute squared Euclidean distance to another vector (faster than distance())."""
        other = to_vector(other)
        diff = self._data - other._data
        return float(np.sum(diff * diff))

    def angle(self, other: VectorLike) -> float:
        """Compute the angle (in radians) between this vector and another."""
        other = to_vector(other)
        if self.length() < 1e-10 or other.length() < 1e-10:
            return 0.0

        cos_angle = np.clip(
            np.dot(self._data, other._data)
            / (np.linalg.norm(self._data) * np.linalg.norm(other._data)),
            -1.0,
            1.0,
        )
        return float(np.arccos(cos_angle))

    def project(self: T, onto: VectorLike) -> T:
        """Project this vector onto another vector."""
        onto = to_vector(onto)
        onto_length_sq = np.sum(onto._data * onto._data)
        if onto_length_sq < 1e-10:
            return self.__class__(np.zeros_like(self._data))

        scalar_projection = np.dot(self._data, onto._data) / onto_length_sq
        return self.__class__(scalar_projection * onto._data)

    @classmethod
    def zeros(cls: type[T], dim: int) -> T:
        """Create a zero vector of given dimension."""
        return cls(np.zeros(dim))

    @classmethod
    def ones(cls: type[T], dim: int) -> T:
        """Create a vector of ones with given dimension."""
        return cls(np.ones(dim))

    @classmethod
    def unit_x(cls: type[T], dim: int = 3) -> T:
        """Create a unit vector in the x direction."""
        v = np.zeros(dim)
        v[0] = 1.0
        return cls(v)

    @classmethod
    def unit_y(cls: type[T], dim: int = 3) -> T:
        """Create a unit vector in the y direction."""
        v = np.zeros(dim)
        v[1] = 1.0
        return cls(v)

    @classmethod
    def unit_z(cls: type[T], dim: int = 3) -> T:
        """Create a unit vector in the z direction."""
        v = np.zeros(dim)
        if dim > 2:
            v[2] = 1.0
        return cls(v)

    def to_list(self) -> list[float]:
        """Convert the vector to a list."""
        return self._data.tolist()  # type: ignore[no-any-return]

    def to_tuple(self) -> builtins.tuple[float, ...]:
        """Convert the vector to a tuple."""
        return tuple(self._data)

    def to_numpy(self) -> np.ndarray:  # type: ignore[type-arg]
        """Convert the vector to a numpy array."""
        return self._data

    def is_zero(self) -> bool:
        """Check if this is a zero vector (all components are zero).

        Returns:
            True if all components are zero, False otherwise
        """
        return np.allclose(self._data, 0.0)

    def __bool__(self) -> bool:
        """Boolean conversion for Vector.

        A Vector is considered False if it's a zero vector (all components are zero),
        and True otherwise.

        Returns:
            False if vector is zero, True otherwise
        """
        return not self.is_zero()


def to_numpy(value: VectorLike) -> np.ndarray:  # type: ignore[type-arg]
    """Convert a vector-compatible value to a numpy array.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        Numpy array representation
    """
    if isinstance(value, Vector3):
        return np.array([value.x, value.y, value.z], dtype=float)
    if isinstance(value, Vector):
        return value.data
    elif isinstance(value, np.ndarray):
        return value
    else:
        return np.array(value, dtype=float)


def to_vector(value: VectorLike) -> Vector:
    """Convert a vector-compatible value to a Vector object.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        Vector object
    """
    if isinstance(value, Vector):
        return value
    else:
        return Vector(value)


def to_tuple(value: VectorLike) -> tuple[float, ...]:
    """Convert a vector-compatible value to a tuple.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        Tuple of floats
    """
    if isinstance(value, Vector3):
        return tuple([value.x, value.y, value.z])
    if isinstance(value, Vector):
        return tuple(value.data)
    elif isinstance(value, np.ndarray):
        return tuple(value.tolist())
    elif isinstance(value, tuple):
        return value
    else:
        return tuple(value)


def to_list(value: VectorLike) -> list[float]:
    """Convert a vector-compatible value to a list.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        List of floats
    """
    if isinstance(value, Vector):
        return value.data.tolist()  # type: ignore[no-any-return]
    elif isinstance(value, np.ndarray):
        return value.tolist()  # type: ignore[no-any-return]
    elif isinstance(value, list):
        return value
    else:
        return list(value)  # type: ignore[arg-type]


# Helper functions to check dimensionality
def is_2d(value: VectorLike) -> bool:
    """Check if a vector-compatible value is 2D.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        True if the value is 2D
    """
    if isinstance(value, Vector3):
        return False
    elif isinstance(value, Vector):
        return len(value) == 2  # type: ignore[arg-type]
    elif isinstance(value, np.ndarray):
        return value.shape[-1] == 2 or value.size == 2
    else:
        return len(value) == 2


def is_3d(value: VectorLike) -> bool:
    """Check if a vector-compatible value is 3D.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        True if the value is 3D
    """
    if isinstance(value, Vector):
        return len(value) == 3  # type: ignore[arg-type]
    elif isinstance(value, Vector3):
        return True
    elif isinstance(value, np.ndarray):
        return value.shape[-1] == 3 or value.size == 3
    else:
        return len(value) == 3


# Extraction functions for XYZ components
def x(value: VectorLike) -> float:
    """Get the X component of a vector-compatible value.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        X component as a float
    """
    if isinstance(value, Vector):
        return value.x
    elif isinstance(value, Vector3):
        return value.x  # type: ignore[no-any-return]
    else:
        return float(to_numpy(value)[0])


def y(value: VectorLike) -> float:
    """Get the Y component of a vector-compatible value.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        Y component as a float
    """
    if isinstance(value, Vector):
        return value.y
    elif isinstance(value, Vector3):
        return value.y  # type: ignore[no-any-return]
    else:
        arr = to_numpy(value)
        return float(arr[1]) if len(arr) > 1 else 0.0


def z(value: VectorLike) -> float:
    """Get the Z component of a vector-compatible value.

    Args:
        value: Any vector-like object (Vector, numpy array, tuple, list)

    Returns:
        Z component as a float
    """
    if isinstance(value, Vector):
        return value.z
    elif isinstance(value, Vector3):
        return value.z  # type: ignore[no-any-return]
    else:
        arr = to_numpy(value)
        return float(arr[2]) if len(arr) > 2 else 0.0



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Union, get_args, get_origin


def python_type_to_json_schema(python_type) -> dict:  # type: ignore[no-untyped-def, type-arg]
    """Convert Python type annotations to JSON Schema format."""
    # Handle None/NoneType
    if python_type is type(None) or python_type is None:
        return {"type": "null"}

    # Handle Union types (including Optional)
    origin = get_origin(python_type)
    if origin is Union:
        args = get_args(python_type)
        # Handle Optional[T] which is Union[T, None]
        if len(args) == 2 and type(None) in args:
            non_none_type = args[0] if args[1] is type(None) else args[1]
            schema = python_type_to_json_schema(non_none_type)
            # For OpenAI function calling, we don't use anyOf for optional params
            return schema
        else:
            # For other Union types, use anyOf
            return {"anyOf": [python_type_to_json_schema(arg) for arg in args]}

    # Handle List/list types
    if origin in (list, list):
        args = get_args(python_type)
        if args:
            return {"type": "array", "items": python_type_to_json_schema(args[0])}
        return {"type": "array"}

    # Handle Dict/dict types
    if origin in (dict, dict):
        return {"type": "object"}

    # Handle basic types
    type_map = {
        str: {"type": "string"},
        int: {"type": "integer"},
        float: {"type": "number"},
        bool: {"type": "boolean"},
        list: {"type": "array"},
        dict: {"type": "object"},
    }

    return type_map.get(python_type, {"type": "string"})


def function_to_schema(func) -> dict:  # type: ignore[no-untyped-def, type-arg]
    """Convert a function to OpenAI function schema format."""
    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(f"Failed to get signature for function {func.__name__}: {e!s}")

    properties = {}
    required = []

    for param_name, param in signature.parameters.items():
        # Skip 'self' parameter for methods
        if param_name == "self":
            continue

        # Get the type annotation
        if param.annotation != inspect.Parameter.empty:
            param_schema = python_type_to_json_schema(param.annotation)
        else:
            # Default to string if no type annotation
            param_schema = {"type": "string"}

        # Add description from docstring if available (would need more sophisticated parsing)
        properties[param_name] = param_schema

        # Add to required list if no default value
        if param.default == inspect.Parameter.empty:
            required.append(param_name)

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": (func.__doc__ or "").strip(),
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": required,
            },
        },
    }



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for RPC exception serialization utilities."""

from dimos.protocol.rpc.rpc_utils import (
    RemoteError,
    deserialize_exception,
    serialize_exception,
)


def test_exception_builtin_serialization():
    """Test serialization and deserialization of exceptions."""

    # Test with a builtin exception
    try:
        raise ValueError("test error", 42)
    except ValueError as e:
        serialized = serialize_exception(e)

        # Check serialized format
        assert serialized["type_name"] == "ValueError"
        assert serialized["type_module"] == "builtins"
        assert serialized["args"] == ("test error", 42)
        assert "Traceback" in serialized["traceback"]
        assert "test error" in serialized["traceback"]

        # Deserialize and check we get a real ValueError back
        deserialized = deserialize_exception(serialized)
        assert isinstance(deserialized, ValueError)
        assert deserialized.args == ("test error", 42)
        # Check that remote traceback is attached as cause
        assert isinstance(deserialized.__cause__, RemoteError)
        assert "test error" in deserialized.__cause__.remote_traceback


def test_exception_custom_serialization():
    # Test with a custom exception
    class CustomError(Exception):
        pass

    try:
        raise CustomError("custom message")
    except CustomError as e:
        serialized = serialize_exception(e)

        # Check serialized format
        assert serialized["type_name"] == "CustomError"
        # Module name varies when running under pytest vs directly
        assert serialized["type_module"] in ("__main__", "dimos.protocol.rpc.test_rpc_utils")
        assert serialized["args"] == ("custom message",)

        # Deserialize - should get RemoteError since it's not builtin
        deserialized = deserialize_exception(serialized)
        assert isinstance(deserialized, RemoteError)
        assert "CustomError" in deserialized.remote_type
        assert "custom message" in str(deserialized)
        assert "custom message" in deserialized.remote_traceback



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from collections.abc import Callable
import time

import pytest

from dimos import core
from dimos.core import Module, rpc
from dimos.msgs.sensor_msgs import PointCloud2
from dimos.robot.unitree_webrtc.type.map import Map as Mapper


@pytest.fixture
def dimos():
    return core.start(2)


@pytest.fixture
def client():
    return core.start(2)


class Consumer:
    testf: Callable[[int], int]

    def __init__(self, counter=None) -> None:
        self.testf = counter
        print("consumer init with", counter)

    async def waitcall(self, n: int):
        async def task() -> None:
            await asyncio.sleep(n)

            print("sleep finished, calling")
            res = await self.testf(n)
            print("res is", res)

        asyncio.create_task(task())
        return n


class Counter(Module):
    @rpc
    def addten(self, x: int):
        print(f"counter adding to {x}")
        return x + 10


@pytest.mark.tool
def test_wait(client) -> None:
    counter = client.submit(Counter, actor=True).result()

    async def addten(n):
        return await counter.addten(n)

    consumer = client.submit(Consumer, counter=addten, actor=True).result()

    print("waitcall1", consumer.waitcall(2).result())
    print("waitcall2", consumer.waitcall(2).result())
    time.sleep(1)


@pytest.mark.tool
def test_basic(dimos) -> None:
    counter = dimos.deploy(Counter)
    consumer = dimos.deploy(
        Consumer,
        counter=lambda x: counter.addten(x).result(),
    )

    print(consumer)
    print(counter)
    print("starting consumer")
    consumer.start().result()

    res = consumer.inc(10).result()

    print("result is", res)
    assert res == 20


@pytest.mark.tool
def test_mapper_start(dimos) -> None:
    mapper = dimos.deploy(Mapper)
    mapper.lidar.transport = core.LCMTransport("/lidar", PointCloud2)
    print("start res", mapper.start().result())


if __name__ == "__main__":
    dimos = core.start(2)
    test_basic(dimos)
    test_mapper_start(dimos)


@pytest.mark.tool
def test_counter(dimos) -> None:
    counter = dimos.deploy(Counter)
    assert counter.addten(10) == 20



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Any

# from dimos.core.core import rpc
from dimos.protocol.skill.comms import LCMSkillComms, SkillCommsSpec
from dimos.protocol.skill.schema import function_to_schema
from dimos.protocol.skill.type import (
    MsgType,
    Output,
    Reducer,
    Return,
    SkillConfig,
    SkillMsg,
    Stream,
)

# skill is a decorator that allows us to specify a skill behaviour for a function.
#
# there are several parameters that can be specified:
# - ret: how to return the value from the skill, can be one of:
#
#        Return.none: doesn't return anything to an agent
#        Return.passive: doesn't schedule an agent call but
#                        returns the value to the agent when agent is called
#        Return.call_agent: calls the agent with the value, scheduling an agent call
#
# - stream: if the skill streams values, it can behave in several ways:
#
#        Stream.none: no streaming, skill doesn't emit any values
#        Stream.passive: doesn't schedule an agent call upon emitting a value,
#                        returns the streamed value to the agent when agent is called
#        Stream.call_agent: calls the agent with every value emitted, scheduling an agent call
#
# - reducer: defines an optional strategy for passive streams and how we collapse potential
#            multiple values into something meaningful for the agent
#
#        Reducer.none: no reduction, every emitted value is returned to the agent
#        Reducer.latest: only the latest value is returned to the agent
#        Reducer.average: assumes the skill emits a number,
#                         the average of all values is returned to the agent


def rpc(fn: Callable[..., Any]) -> Callable[..., Any]:
    fn.__rpc__ = True  # type: ignore[attr-defined]
    return fn


def skill(
    reducer: Reducer = Reducer.latest,  # type: ignore[assignment]
    stream: Stream = Stream.none,
    ret: Return = Return.call_agent,
    output: Output = Output.standard,
    hide_skill: bool = False,
) -> Callable:  # type: ignore[type-arg]
    def decorator(f: Callable[..., Any]) -> Any:
        def wrapper(self, *args, **kwargs):  # type: ignore[no-untyped-def]
            skill = f"{f.__name__}"

            call_id = kwargs.get("call_id", None)
            if call_id:
                del kwargs["call_id"]

                return self.call_skill(call_id, skill, args, kwargs)
                # def run_function():
                #    return self.call_skill(call_id, skill, args, kwargs)
                #
                # thread = threading.Thread(target=run_function)
                # thread.start()
                # return None

            return f(self, *args, **kwargs)

        # sig = inspect.signature(f)
        # params = list(sig.parameters.values())
        # if params and params[0].name == "self":
        #     params = params[1:]  # Remove first parameter 'self'
        # wrapper.__signature__ = sig.replace(parameters=params)

        skill_config = SkillConfig(
            name=f.__name__,
            reducer=reducer,  # type: ignore[arg-type]
            stream=stream,
            # if stream is passive, ret must be passive too
            ret=ret.passive if stream == Stream.passive else ret,
            output=output,
            schema=function_to_schema(f),
            hide_skill=hide_skill,
        )

        wrapper.__rpc__ = True  # type: ignore[attr-defined]
        wrapper._skill_config = skill_config  # type: ignore[attr-defined]
        wrapper.__name__ = f.__name__  # Preserve original function name
        wrapper.__doc__ = f.__doc__  # Preserve original docstring
        return wrapper

    return decorator


@dataclass
class SkillContainerConfig:
    skill_transport: type[SkillCommsSpec] = LCMSkillComms


def threaded(f: Callable[..., Any]) -> Callable[..., None]:
    """Decorator to run a function in a thread pool."""

    def wrapper(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        if self._skill_thread_pool is None:
            self._skill_thread_pool = ThreadPoolExecutor(
                max_workers=50, thread_name_prefix="skill_worker"
            )
        self._skill_thread_pool.submit(f, self, *args, **kwargs)
        return None

    return wrapper


# Inherited by any class that wants to provide skills
# (This component works standalone but commonly used by DimOS modules)
#
# Hosts the function execution and handles correct publishing of skill messages
# according to the individual skill decorator configuration
#
# - It allows us to specify a communication layer for skills (LCM for now by default)
# - introspection of available skills via the `skills` RPC method
# - ability to provide dynamic context dependant skills with dynamic_skills flag
#   for this you'll need to override the `skills` method to return a dynamic set of skills
#   SkillCoordinator will call this method to get the skills available upon every request to
#   the agent


class SkillContainer:
    skill_transport_class: type[SkillCommsSpec] = LCMSkillComms
    _skill_thread_pool: ThreadPoolExecutor | None = None
    _skill_transport: SkillCommsSpec | None = None

    @rpc
    def dynamic_skills(self) -> bool:
        return False

    def __str__(self) -> str:
        return f"SkillContainer({self.__class__.__name__})"

    @rpc
    def stop(self) -> None:
        if self._skill_transport:
            self._skill_transport.stop()
            self._skill_transport = None

        if self._skill_thread_pool:
            self._skill_thread_pool.shutdown(wait=True)
            self._skill_thread_pool = None

        # Continue the MRO chain if there's a parent stop() method
        if hasattr(super(), "stop"):
            super().stop()  # type: ignore[misc]

    # TODO: figure out standard args/kwargs passing format,
    # use same interface as skill coordinator call_skill method
    @threaded
    def call_skill(
        self, call_id: str, skill_name: str, args: tuple[Any, ...], kwargs: dict[str, Any]
    ) -> None:
        f = getattr(self, skill_name, None)

        if f is None:
            raise ValueError(f"Function '{skill_name}' not found in {self.__class__.__name__}")

        config = getattr(f, "_skill_config", None)
        if config is None:
            raise ValueError(f"Function '{skill_name}' in {self.__class__.__name__} is not a skill")

        # we notify the skill transport about the start of the skill call
        self.skill_transport.publish(SkillMsg(call_id, skill_name, None, type=MsgType.start))

        try:
            val = f(*args, **kwargs)

            # check if the skill returned a coroutine, if it is, block until it resolves
            if isinstance(val, asyncio.Future):
                val = asyncio.run(val)  # type: ignore[arg-type]

            # check if the skill is a generator, if it is, we need to iterate over it
            if hasattr(val, "__iter__") and not isinstance(val, str):
                last_value = None
                for v in val:
                    last_value = v
                    self.skill_transport.publish(
                        SkillMsg(call_id, skill_name, v, type=MsgType.stream)
                    )
                self.skill_transport.publish(
                    SkillMsg(call_id, skill_name, last_value, type=MsgType.ret)
                )

            else:
                self.skill_transport.publish(SkillMsg(call_id, skill_name, val, type=MsgType.ret))

        except Exception as e:
            import traceback

            formatted_traceback = "".join(traceback.TracebackException.from_exception(e).format())

            self.skill_transport.publish(
                SkillMsg(
                    call_id,
                    skill_name,
                    {"msg": str(e), "traceback": formatted_traceback},
                    type=MsgType.error,
                )
            )

    @rpc
    def skills(self) -> dict[str, SkillConfig]:
        # Avoid recursion by excluding this property itself
        # Also exclude known properties that shouldn't be accessed
        excluded = {"skills", "tf", "rpc", "skill_transport"}
        return {
            name: getattr(self, name)._skill_config
            for name in dir(self)
            if not name.startswith("_")
            and name not in excluded
            and hasattr(getattr(self, name), "_skill_config")
        }

    @property
    def skill_transport(self) -> SkillCommsSpec:
        if self._skill_transport is None:
            self._skill_transport = self.skill_transport_class()
        return self._skill_transport



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract base class for manipulation skills."""

from dimos.manipulation.manipulation_interface import ManipulationInterface
from dimos.robot.robot import Robot
from dimos.skills.skills import AbstractRobotSkill
from dimos.types.robot_capabilities import RobotCapability


class AbstractManipulationSkill(AbstractRobotSkill):
    """Base class for all manipulation-related skills.

    This abstract class provides access to the robot's manipulation memory system.
    """

    def __init__(self, *args, robot: Robot | None = None, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Initialize the manipulation skill.

        Args:
            robot: The robot instance to associate with this skill
        """
        super().__init__(*args, robot=robot, **kwargs)

        if self._robot and not self._robot.manipulation_interface:  # type: ignore[attr-defined]
            raise NotImplementedError(
                "This robot does not have a manipulation interface implemented"
            )

    @property
    def manipulation_interface(self) -> ManipulationInterface | None:
        """Get the robot's manipulation interface.

        Returns:
            ManipulationInterface: The robot's manipulation interface or None if not available

        Raises:
            RuntimeError: If the robot doesn't have the MANIPULATION capability
        """
        if self._robot is None:
            return None

        if not self._robot.has_capability(RobotCapability.MANIPULATION):
            raise RuntimeError("This robot does not have manipulation capabilities")

        return self._robot.manipulation_interface  # type: ignore[attr-defined, no-any-return]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pick and place skill for Piper Arm robot.

This module provides a skill that uses Qwen VLM to identify pick and place
locations based on natural language queries, then executes the manipulation.
"""

import json
import os
from typing import Any

import cv2
import numpy as np
from pydantic import Field

from dimos.models.qwen.video_query import query_single_frame
from dimos.skills.skills import AbstractRobotSkill
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


def parse_qwen_points_response(response: str) -> tuple[tuple[int, int], tuple[int, int]] | None:
    """
    Parse Qwen's response containing two points.

    Args:
        response: Qwen's response containing JSON with two points

    Returns:
        Tuple of (pick_point, place_point) where each point is (x, y), or None if parsing fails
    """
    try:
        # Try to extract JSON from the response
        start_idx = response.find("{")
        end_idx = response.rfind("}") + 1

        if start_idx >= 0 and end_idx > start_idx:
            json_str = response[start_idx:end_idx]
            result = json.loads(json_str)

            # Extract pick and place points
            if "pick_point" in result and "place_point" in result:
                pick = result["pick_point"]
                place = result["place_point"]

                # Validate points have x,y coordinates
                if (
                    isinstance(pick, list | tuple)
                    and len(pick) >= 2
                    and isinstance(place, list | tuple)
                    and len(place) >= 2
                ):
                    return (int(pick[0]), int(pick[1])), (int(place[0]), int(place[1]))

    except Exception as e:
        logger.error(f"Error parsing Qwen points response: {e}")
        logger.debug(f"Raw response: {response}")

    return None


def save_debug_image_with_points(
    image: np.ndarray,  # type: ignore[type-arg]
    pick_point: tuple[int, int] | None = None,
    place_point: tuple[int, int] | None = None,
    filename_prefix: str = "qwen_debug",
) -> str:
    """
    Save debug image with crosshairs marking pick and/or place points.

    Args:
        image: RGB image array
        pick_point: (x, y) coordinates for pick location
        place_point: (x, y) coordinates for place location
        filename_prefix: Prefix for the saved filename

    Returns:
        Path to the saved image
    """
    # Create a copy to avoid modifying original
    debug_image = image.copy()

    # Draw pick point crosshair (green)
    if pick_point:
        x, y = pick_point
        # Draw crosshair
        cv2.drawMarker(debug_image, (x, y), (0, 255, 0), cv2.MARKER_CROSS, 30, 2)
        # Draw circle
        cv2.circle(debug_image, (x, y), 5, (0, 255, 0), -1)
        # Add label
        cv2.putText(
            debug_image, "PICK", (x + 10, y - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 255, 0), 2
        )

    # Draw place point crosshair (cyan)
    if place_point:
        x, y = place_point
        # Draw crosshair
        cv2.drawMarker(debug_image, (x, y), (255, 255, 0), cv2.MARKER_CROSS, 30, 2)
        # Draw circle
        cv2.circle(debug_image, (x, y), 5, (255, 255, 0), -1)
        # Add label
        cv2.putText(
            debug_image, "PLACE", (x + 10, y - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.7, (255, 255, 0), 2
        )

    # Draw arrow from pick to place if both exist
    if pick_point and place_point:
        cv2.arrowedLine(debug_image, pick_point, place_point, (255, 0, 255), 2, tipLength=0.03)

    # Generate filename with timestamp
    filename = f"{filename_prefix}.png"
    filepath = os.path.join(os.getcwd(), filename)

    # Save image
    cv2.imwrite(filepath, debug_image)
    logger.info(f"Debug image saved to: {filepath}")

    return filepath


def parse_qwen_single_point_response(response: str) -> tuple[int, int] | None:
    """
    Parse Qwen's response containing a single point.

    Args:
        response: Qwen's response containing JSON with a point

    Returns:
        Tuple of (x, y) or None if parsing fails
    """
    try:
        # Try to extract JSON from the response
        start_idx = response.find("{")
        end_idx = response.rfind("}") + 1

        if start_idx >= 0 and end_idx > start_idx:
            json_str = response[start_idx:end_idx]
            result = json.loads(json_str)

            # Try different possible keys
            point = None
            for key in ["point", "location", "position", "coordinates"]:
                if key in result:
                    point = result[key]
                    break

            # Validate point has x,y coordinates
            if point and isinstance(point, list | tuple) and len(point) >= 2:
                return int(point[0]), int(point[1])

    except Exception as e:
        logger.error(f"Error parsing Qwen single point response: {e}")
        logger.debug(f"Raw response: {response}")

    return None


class PickAndPlace(AbstractRobotSkill):
    """
    A skill that performs pick and place operations using vision-language guidance.

    This skill uses Qwen VLM to identify objects and locations based on natural
    language queries, then executes pick and place operations using the robot's
    manipulation interface.

    Example usage:
        # Just pick the object
        skill = PickAndPlace(robot=robot, object_query="red mug")

        # Pick and place the object
        skill = PickAndPlace(robot=robot, object_query="red mug", target_query="on the coaster")

    The skill uses the robot's stereo camera to capture RGB images and its manipulation
    interface to execute the pick and place operation. It automatically handles coordinate
    transformation from 2D pixel coordinates to 3D world coordinates.
    """

    object_query: str = Field(
        "mug",
        description="Natural language description of the object to pick (e.g., 'red mug', 'small box')",
    )

    target_query: str | None = Field(
        None,
        description="Natural language description of where to place the object (e.g., 'on the table', 'in the basket'). If not provided, only pick operation will be performed.",
    )

    model_name: str = Field(
        "qwen2.5-vl-72b-instruct", description="Qwen model to use for visual queries"
    )

    def __init__(self, robot=None, **data) -> None:  # type: ignore[no-untyped-def]
        """
        Initialize the PickAndPlace skill.

        Args:
            robot: The PiperArmRobot instance
            **data: Additional configuration data
        """
        super().__init__(robot=robot, **data)

    def _get_camera_frame(self) -> np.ndarray | None:  # type: ignore[type-arg]
        """
        Get a single RGB frame from the robot's camera.

        Returns:
            RGB image as numpy array or None if capture fails
        """
        if not self._robot or not self._robot.manipulation_interface:  # type: ignore[attr-defined]
            logger.error("Robot or stereo camera not available")
            return None

        try:
            # Use the RPC call to get a single RGB frame
            rgb_frame = self._robot.manipulation_interface.get_single_rgb_frame()  # type: ignore[attr-defined]
            if rgb_frame is None:
                logger.error("Failed to capture RGB frame from camera")
            return rgb_frame  # type: ignore[no-any-return]
        except Exception as e:
            logger.error(f"Error getting camera frame: {e}")
            return None

    def _query_pick_and_place_points(
        self,
        frame: np.ndarray,  # type: ignore[type-arg]
    ) -> tuple[tuple[int, int], tuple[int, int]] | None:
        """
        Query Qwen to get both pick and place points in a single query.

        Args:
            frame: RGB image array

        Returns:
            Tuple of (pick_point, place_point) or None if query fails
        """
        # This method is only called when both object and target are specified
        prompt = (
            f"Look at this image carefully. I need you to identify two specific locations:\n"
            f"1. Find the {self.object_query} - this is the object I want to pick up\n"
            f"2. Identify where to place it {self.target_query}\n\n"
            "Instructions:\n"
            "- The pick_point should be at the center or graspable part of the object\n"
            "- The place_point should be a stable, flat surface at the target location\n"
            "- Consider the object's size when choosing the placement point\n\n"
            "Return ONLY a JSON object with this exact format:\n"
            "{'pick_point': [x, y], 'place_point': [x, y]}\n"
            "where [x, y] are pixel coordinates in the image."
        )

        try:
            response = query_single_frame(frame, prompt, model_name=self.model_name)
            return parse_qwen_points_response(response)
        except Exception as e:
            logger.error(f"Error querying Qwen for pick and place points: {e}")
            return None

    def _query_single_point(
        self,
        frame: np.ndarray,  # type: ignore[type-arg]
        query: str,
        point_type: str,
    ) -> tuple[int, int] | None:
        """
        Query Qwen to get a single point location.

        Args:
            frame: RGB image array
            query: Natural language description of what to find
            point_type: Type of point ('pick' or 'place') for context

        Returns:
            Tuple of (x, y) pixel coordinates or None if query fails
        """
        if point_type == "pick":
            prompt = (
                f"Look at this image carefully and find the {query}.\n\n"
                "Instructions:\n"
                "- Identify the exact object matching the description\n"
                "- Choose the center point or the most graspable location on the object\n"
                "- If multiple matching objects exist, choose the most prominent or accessible one\n"
                "- Consider the object's shape and material when selecting the grasp point\n\n"
                "Return ONLY a JSON object with this exact format:\n"
                "{'point': [x, y]}\n"
                "where [x, y] are the pixel coordinates of the optimal grasping point on the object."
            )
        else:  # place
            prompt = (
                f"Look at this image and identify where to place an object {query}.\n\n"
                "Instructions:\n"
                "- Find a stable, flat surface at the specified location\n"
                "- Ensure the placement spot is clear of obstacles\n"
                "- Consider the size of the object being placed\n"
                "- If the query specifies a container or specific spot, center the placement there\n"
                "- Otherwise, find the most appropriate nearby surface\n\n"
                "Return ONLY a JSON object with this exact format:\n"
                "{'point': [x, y]}\n"
                "where [x, y] are the pixel coordinates of the optimal placement location."
            )

        try:
            response = query_single_frame(frame, prompt, model_name=self.model_name)
            return parse_qwen_single_point_response(response)
        except Exception as e:
            logger.error(f"Error querying Qwen for {point_type} point: {e}")
            return None

    def __call__(self) -> dict[str, Any]:
        """
        Execute the pick and place operation.

        Returns:
            Dictionary with operation results
        """
        super().__call__()  # type: ignore[no-untyped-call]

        if not self._robot:
            error_msg = "No robot instance provided to PickAndPlace skill"
            logger.error(error_msg)
            return {"success": False, "error": error_msg}

        # Register skill as running
        skill_library = self._robot.get_skills()  # type: ignore[no-untyped-call]
        self.register_as_running("PickAndPlace", skill_library)

        # Get camera frame
        frame = self._get_camera_frame()
        if frame is None:
            return {"success": False, "error": "Failed to capture camera frame"}

        # Convert RGB to BGR for OpenCV if needed
        if len(frame.shape) == 3 and frame.shape[2] == 3:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

        # Get pick and place points from Qwen
        pick_point = None
        place_point = None

        # Determine mode based on whether target_query is provided
        if self.target_query is None:
            # Pick only mode
            logger.info("Pick-only mode (no target specified)")

            # Query for pick point
            pick_point = self._query_single_point(frame, self.object_query, "pick")
            if not pick_point:
                return {"success": False, "error": f"Failed to find {self.object_query}"}

            # No place point needed for pick-only
            place_point = None
        else:
            # Pick and place mode - can use either single or dual query
            logger.info("Pick and place mode (target specified)")

            # Try single query first for efficiency
            points = self._query_pick_and_place_points(frame)
            pick_point, place_point = points  # type: ignore[misc]

        logger.info(f"Pick point: {pick_point}, Place point: {place_point}")

        # Save debug image with marked points
        if pick_point or place_point:
            save_debug_image_with_points(frame, pick_point, place_point)

        # Execute pick (and optionally place) using the robot's interface
        try:
            if place_point:
                # Pick and place
                result = self._robot.pick_and_place(  # type: ignore[attr-defined]
                    pick_x=pick_point[0],
                    pick_y=pick_point[1],
                    place_x=place_point[0],
                    place_y=place_point[1],
                )
            else:
                # Pick only
                result = self._robot.pick_and_place(  # type: ignore[attr-defined]
                    pick_x=pick_point[0], pick_y=pick_point[1], place_x=None, place_y=None
                )

            if result:
                if self.target_query:
                    message = (
                        f"Successfully picked {self.object_query} and placed it {self.target_query}"
                    )
                else:
                    message = f"Successfully picked {self.object_query}"

                return {
                    "success": True,
                    "pick_point": pick_point,
                    "place_point": place_point,
                    "object": self.object_query,
                    "target": self.target_query,
                    "message": message,
                }
            else:
                operation = "Pick and place" if self.target_query else "Pick"
                return {
                    "success": False,
                    "pick_point": pick_point,
                    "place_point": place_point,
                    "error": f"{operation} operation failed",
                }

        except Exception as e:
            logger.error(f"Error executing pick and place: {e}")
            return {
                "success": False,
                "error": f"Execution error: {e!s}",
                "pick_point": pick_point,
                "place_point": place_point,
            }
        finally:
            # Always unregister skill when done
            self.stop()

    def stop(self) -> None:
        """
        Stop the pick and place operation and perform cleanup.
        """
        logger.info("Stopping PickAndPlace skill")

        # Unregister skill from skill library
        if self._robot:
            skill_library = self._robot.get_skills()  # type: ignore[no-untyped-call]
            self.unregister_as_running("PickAndPlace", skill_library)

        logger.info("PickAndPlace skill stopped successfully")



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from enum import Enum

from dimos.msgs.geometry_msgs import PoseStamped


class NavigationState(Enum):
    IDLE = "idle"
    FOLLOWING_PATH = "following_path"
    RECOVERY = "recovery"


class NavigationInterface(ABC):
    @abstractmethod
    def set_goal(self, goal: PoseStamped) -> bool:
        """
        Set a new navigation goal (non-blocking).

        Args:
            goal: Target pose to navigate to

        Returns:
            True if goal was accepted, False otherwise
        """
        pass

    @abstractmethod
    def get_state(self) -> NavigationState:
        """
        Get the current state of the navigator.

        Returns:
            Current navigation state
        """
        pass

    @abstractmethod
    def is_goal_reached(self) -> bool:
        """
        Check if the current goal has been reached.

        Returns:
            True if goal was reached, False otherwise
        """
        pass

    @abstractmethod
    def cancel_goal(self) -> bool:
        """
        Cancel the current navigation goal.

        Returns:
            True if goal was cancelled, False if no goal was active
        """
        pass


__all__ = ["NavigationInterface", "NavigationState"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Any

from dimos.core.core import rpc
from dimos.core.skill_module import SkillModule
from dimos.core.stream import In
from dimos.models.qwen.video_query import BBox
from dimos.models.vl.qwen import QwenVlModel
from dimos.msgs.geometry_msgs import PoseStamped, Quaternion, Vector3
from dimos.msgs.geometry_msgs.Vector3 import make_vector3
from dimos.msgs.sensor_msgs import Image
from dimos.navigation.base import NavigationState
from dimos.navigation.visual.query import get_object_bbox_from_image
from dimos.protocol.skill.skill import skill
from dimos.types.robot_location import RobotLocation
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class NavigationSkillContainer(SkillModule):
    _latest_image: Image | None = None
    _latest_odom: PoseStamped | None = None
    _skill_started: bool = False
    _similarity_threshold: float = 0.23

    rpc_calls: list[str] = [
        "SpatialMemory.tag_location",
        "SpatialMemory.query_tagged_location",
        "SpatialMemory.query_by_text",
        "NavigationInterface.set_goal",
        "NavigationInterface.get_state",
        "NavigationInterface.is_goal_reached",
        "NavigationInterface.cancel_goal",
        "ObjectTracking.track",
        "ObjectTracking.stop_track",
        "ObjectTracking.is_tracking",
        "WavefrontFrontierExplorer.stop_exploration",
        "WavefrontFrontierExplorer.explore",
        "WavefrontFrontierExplorer.is_exploration_active",
    ]

    color_image: In[Image]
    odom: In[PoseStamped]

    def __init__(self) -> None:
        super().__init__()
        self._skill_started = False
        self._vl_model = QwenVlModel()

    @rpc
    def start(self) -> None:
        self._disposables.add(self.color_image.subscribe(self._on_color_image))  # type: ignore[arg-type]
        self._disposables.add(self.odom.subscribe(self._on_odom))  # type: ignore[arg-type]
        self._skill_started = True

    @rpc
    def stop(self) -> None:
        super().stop()

    def _on_color_image(self, image: Image) -> None:
        self._latest_image = image

    def _on_odom(self, odom: PoseStamped) -> None:
        self._latest_odom = odom

    @skill()
    def tag_location(self, location_name: str) -> str:
        """Tag this location in the spatial memory with a name.

        This associates the current location with the given name in the spatial memory, allowing you to navigate back to it.

        Args:
            location_name (str): the name for the location

        Returns:
            str: the outcome
        """

        if not self._skill_started:
            raise ValueError(f"{self} has not been started.")
        tf = self.tf.get("map", "base_link", time_tolerance=2.0)
        if not tf:
            return "Could not get the robot's current transform."

        position = tf.translation
        rotation = tf.rotation.to_euler()

        location = RobotLocation(
            name=location_name,
            position=(position.x, position.y, position.z),
            rotation=(rotation.x, rotation.y, rotation.z),
        )

        tag_location_rpc = self.get_rpc_calls("SpatialMemory.tag_location")
        if not tag_location_rpc(location):
            return f"Error: Failed to store '{location_name}' in the spatial memory"

        logger.info(f"Tagged {location}")
        return f"Tagged '{location_name}': ({position.x},{position.y})."

    @skill()
    def navigate_with_text(self, query: str) -> str:
        """Navigate to a location by querying the existing semantic map using natural language.

        First attempts to locate an object in the robot's camera view using vision.
        If the object is found, navigates to it. If not, falls back to querying the
        semantic map for a location matching the description.
        CALL THIS SKILL FOR ONE SUBJECT AT A TIME. For example: "Go to the person wearing a blue shirt in the living room",
        you should call this skill twice, once for the person wearing a blue shirt and once for the living room.
        Args:
            query: Text query to search for in the semantic map
        """

        if not self._skill_started:
            raise ValueError(f"{self} has not been started.")
        success_msg = self._navigate_by_tagged_location(query)
        if success_msg:
            return success_msg

        logger.info(f"No tagged location found for {query}")

        success_msg = self._navigate_to_object(query)
        if success_msg:
            return success_msg

        logger.info(f"No object in view found for {query}")

        success_msg = self._navigate_using_semantic_map(query)
        if success_msg:
            return success_msg

        return f"No tagged location called '{query}'. No object in view matching '{query}'. No matching location found in semantic map for '{query}'."

    def _navigate_by_tagged_location(self, query: str) -> str | None:
        try:
            query_tagged_location_rpc = self.get_rpc_calls("SpatialMemory.query_tagged_location")
        except Exception:
            logger.warning("SpatialMemory module not connected, cannot query tagged locations")
            return None

        robot_location = query_tagged_location_rpc(query)

        if not robot_location:
            return None

        print("Found tagged location:", robot_location)
        goal_pose = PoseStamped(
            position=make_vector3(*robot_location.position),
            orientation=Quaternion.from_euler(Vector3(*robot_location.rotation)),
            frame_id="map",
        )

        result = self._navigate_to(goal_pose)
        if not result:
            return "Error: Faild to reach the tagged location."

        return (
            f"Successfuly arrived at location tagged '{robot_location.name}' from query '{query}'."
        )

    def _navigate_to(self, pose: PoseStamped) -> bool:
        try:
            set_goal_rpc, get_state_rpc, is_goal_reached_rpc = self.get_rpc_calls(
                "NavigationInterface.set_goal",
                "NavigationInterface.get_state",
                "NavigationInterface.is_goal_reached",
            )
        except Exception:
            logger.error("Navigation module not connected properly")
            return False

        logger.info(
            f"Navigating to pose: ({pose.position.x:.2f}, {pose.position.y:.2f}, {pose.position.z:.2f})"
        )
        set_goal_rpc(pose)
        time.sleep(1.0)

        while get_state_rpc() == NavigationState.FOLLOWING_PATH:
            time.sleep(0.25)

        time.sleep(1.0)
        if not is_goal_reached_rpc():
            logger.info("Navigation was cancelled or failed")
            return False
        else:
            logger.info("Navigation goal reached")
            return True

    def _navigate_to_object(self, query: str) -> str | None:
        try:
            bbox = self._get_bbox_for_current_frame(query)
        except Exception:
            logger.error(f"Failed to get bbox for {query}", exc_info=True)
            return None

        if bbox is None:
            return None

        try:
            track_rpc, stop_track_rpc, is_tracking_rpc = self.get_rpc_calls(
                "ObjectTracking.track", "ObjectTracking.stop_track", "ObjectTracking.is_tracking"
            )
        except Exception:
            logger.error("ObjectTracking module not connected properly")
            return None

        try:
            get_state_rpc, is_goal_reached_rpc = self.get_rpc_calls(
                "NavigationInterface.get_state", "NavigationInterface.is_goal_reached"
            )
        except Exception:
            logger.error("Navigation module not connected properly")
            return None

        logger.info(f"Found {query} at {bbox}")

        # Start tracking - BBoxNavigationModule automatically generates goals
        track_rpc(bbox)

        start_time = time.time()
        timeout = 30.0
        goal_set = False

        while time.time() - start_time < timeout:
            # Check if navigator finished
            if get_state_rpc() == NavigationState.IDLE and goal_set:
                logger.info("Waiting for goal result")
                time.sleep(1.0)
                if not is_goal_reached_rpc():
                    logger.info(f"Goal cancelled, tracking '{query}' failed")
                    stop_track_rpc()
                    return None
                else:
                    logger.info(f"Reached '{query}'")
                    stop_track_rpc()
                    return f"Successfully arrived at '{query}'"

            # If goal set and tracking lost, just continue (tracker will resume or timeout)
            if goal_set and not is_tracking_rpc():
                continue

            # BBoxNavigationModule automatically sends goals when tracker publishes
            # Just check if we have any detections to mark goal_set
            if is_tracking_rpc():
                goal_set = True

            time.sleep(0.25)

        logger.warning(f"Navigation to '{query}' timed out after {timeout}s")
        stop_track_rpc()
        return None

    def _get_bbox_for_current_frame(self, query: str) -> BBox | None:
        if self._latest_image is None:
            return None

        return get_object_bbox_from_image(self._vl_model, self._latest_image, query)

    def _navigate_using_semantic_map(self, query: str) -> str:
        try:
            query_by_text_rpc = self.get_rpc_calls("SpatialMemory.query_by_text")
        except Exception:
            return "Error: The SpatialMemory module is not connected."

        results = query_by_text_rpc(query)

        if not results:
            return f"No matching location found in semantic map for '{query}'"

        best_match = results[0]

        goal_pose = self._get_goal_pose_from_result(best_match)

        print("Goal pose for semantic nav:", goal_pose)
        if not goal_pose:
            return f"Found a result for '{query}' but it didn't have a valid position."

        result = self._navigate_to(goal_pose)

        if not result:
            return f"Failed to navigate for '{query}'"

        return f"Successfuly arrived at '{query}'"

    @skill()
    def follow_human(self, person: str) -> str:
        """Follow a specific person"""
        return "Not implemented yet."

    @skill()
    def stop_movement(self) -> str:
        """Immediatly stop moving."""

        if not self._skill_started:
            raise ValueError(f"{self} has not been started.")

        self._cancel_goal_and_stop()

        return "Stopped"

    def _cancel_goal_and_stop(self) -> None:
        try:
            cancel_goal_rpc = self.get_rpc_calls("NavigationInterface.cancel_goal")
        except Exception:
            logger.warning("Navigation module not connected, cannot cancel goal")
            return

        try:
            stop_exploration_rpc = self.get_rpc_calls("WavefrontFrontierExplorer.stop_exploration")
        except Exception:
            logger.warning("FrontierExplorer module not connected, cannot stop exploration")
            return

        cancel_goal_rpc()
        return stop_exploration_rpc()  # type: ignore[no-any-return]

    @skill()
    def start_exploration(self, timeout: float = 240.0) -> str:
        """A skill that performs autonomous frontier exploration.

        This skill continuously finds and navigates to unknown frontiers in the environment
        until no more frontiers are found or the exploration is stopped.

        Don't call any other skills except stop_movement skill when needed.

        Args:
            timeout (float, optional): Maximum time (in seconds) allowed for exploration
        """

        if not self._skill_started:
            raise ValueError(f"{self} has not been started.")

        try:
            return self._start_exploration(timeout)
        finally:
            self._cancel_goal_and_stop()

    def _start_exploration(self, timeout: float) -> str:
        try:
            explore_rpc, is_exploration_active_rpc = self.get_rpc_calls(
                "WavefrontFrontierExplorer.explore",
                "WavefrontFrontierExplorer.is_exploration_active",
            )
        except Exception:
            return "Error: The WavefrontFrontierExplorer module is not connected."

        logger.info("Starting autonomous frontier exploration")

        start_time = time.time()

        has_started = explore_rpc()
        if not has_started:
            return "Error: Could not start exploration."

        while time.time() - start_time < timeout and is_exploration_active_rpc():
            time.sleep(0.5)

        return "Exploration completed successfuly"

    def _get_goal_pose_from_result(self, result: dict[str, Any]) -> PoseStamped | None:
        similarity = 1.0 - (result.get("distance") or 1)
        if similarity < self._similarity_threshold:
            logger.warning(
                f"Match found but similarity score ({similarity:.4f}) is below threshold ({self._similarity_threshold})"
            )
            return None

        metadata = result.get("metadata")
        if not metadata:
            return None
        print(metadata)
        first = metadata[0]
        print(first)
        pos_x = first.get("pos_x", 0)
        pos_y = first.get("pos_y", 0)
        theta = first.get("rot_z", 0)

        return PoseStamped(
            position=make_vector3(pos_x, pos_y, 0),
            orientation=Quaternion.from_euler(make_vector3(0, 0, theta)),
            frame_id="map",
        )


navigation_skill = NavigationSkillContainer.blueprint

__all__ = ["NavigationSkillContainer", "navigation_skill"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cached_property

from PIL import Image as PILImage
import torch
from transformers import AutoModelForCausalLM, AutoProcessor  # type: ignore[import-untyped]

from dimos.models.base import HuggingFaceModel
from dimos.models.vl.base import Captioner
from dimos.msgs.sensor_msgs import Image


class Florence2Model(HuggingFaceModel, Captioner):
    """Florence-2 captioning model from Microsoft.

    A lightweight, fast captioning model optimized for generating image descriptions
    without requiring a text prompt. Supports multiple caption detail levels.
    """

    _model_class = AutoModelForCausalLM

    def __init__(
        self,
        model_name: str = "microsoft/Florence-2-base",
        **kwargs: object,
    ) -> None:
        """Initialize Florence-2 model.

        Args:
            model_name: HuggingFace model name. Options:
                - "microsoft/Florence-2-base" (~0.2B, fastest)
                - "microsoft/Florence-2-large" (~0.8B, better quality)
            **kwargs: Additional config options (device, dtype, warmup, etc.)
        """
        super().__init__(model_name=model_name, **kwargs)

    @cached_property
    def _processor(self) -> AutoProcessor:
        return AutoProcessor.from_pretrained(
            self.config.model_name, trust_remote_code=self.config.trust_remote_code
        )

    def caption(self, image: Image, detail: str = "normal") -> str:
        """Generate a caption for the image.

        Args:
            image: Input image to caption
            detail: Level of detail for caption:
                - "brief": Short, concise caption
                - "normal": Standard caption (default)
                - "detailed": More detailed description

        Returns:
            Text description of the image
        """
        # Map detail level to Florence-2 task prompts
        task_prompts = {
            "brief": "<CAPTION>",
            "normal": "<CAPTION>",
            "detailed": "<DETAILED_CAPTION>",
            "more_detailed": "<MORE_DETAILED_CAPTION>",
        }
        task_prompt = task_prompts.get(detail, "<CAPTION>")

        # Convert to PIL
        pil_image = PILImage.fromarray(image.to_rgb().data)

        # Process inputs
        inputs = self._processor(text=task_prompt, images=pil_image, return_tensors="pt")
        inputs = self._move_inputs_to_device(inputs)

        # Generate
        with torch.inference_mode():
            generated_ids = self._model.generate(
                **inputs,
                max_new_tokens=256,
                num_beams=3,
                do_sample=False,
            )

        # Decode
        generated_text = self._processor.batch_decode(generated_ids, skip_special_tokens=False)[0]

        # Parse output - Florence returns structured output
        parsed = self._processor.post_process_generation(
            generated_text, task=task_prompt, image_size=pil_image.size
        )

        # Extract caption from parsed output
        caption: str = parsed.get(task_prompt, generated_text)
        return caption.strip()

    def caption_batch(self, *images: Image) -> list[str]:
        """Generate captions for multiple images efficiently.

        Args:
            images: Input images to caption

        Returns:
            List of text descriptions
        """
        if not images:
            return []

        task_prompt = "<CAPTION>"

        # Convert all to PIL
        pil_images = [PILImage.fromarray(img.to_rgb().data) for img in images]

        # Process batch
        inputs = self._processor(
            text=[task_prompt] * len(images), images=pil_images, return_tensors="pt", padding=True
        )
        inputs = self._move_inputs_to_device(inputs)

        # Generate
        with torch.inference_mode():
            generated_ids = self._model.generate(
                **inputs,
                max_new_tokens=256,
                num_beams=3,
                do_sample=False,
            )

        # Decode all
        generated_texts = self._processor.batch_decode(generated_ids, skip_special_tokens=False)

        # Parse outputs
        captions = []
        for text, pil_img in zip(generated_texts, pil_images, strict=True):
            parsed = self._processor.post_process_generation(
                text, task=task_prompt, image_size=pil_img.size
            )
            captions.append(parsed.get(task_prompt, text).strip())

        return captions

    def start(self) -> None:
        """Start the model with a dummy forward pass."""
        # Load model and processor via base class
        super().start()

        # Run a small inference
        dummy = PILImage.new("RGB", (224, 224), color="gray")
        inputs = self._processor(text="<CAPTION>", images=dummy, return_tensors="pt")
        inputs = self._move_inputs_to_device(inputs)

        with torch.inference_mode():
            self._model.generate(**inputs, max_new_tokens=10)

    def stop(self) -> None:
        """Release model and free GPU memory."""
        # Clean up processor cached property
        if "_processor" in self.__dict__:
            del self.__dict__["_processor"]
        # Call parent which handles _model cleanup
        super().stop()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graph database utility functions for temporal memory."""

import re
from typing import TYPE_CHECKING, Any

from dimos.utils.logging_config import setup_logger

if TYPE_CHECKING:
    from dimos.models.vl.base import VlModel
    from dimos.msgs.sensor_msgs import Image

    from ..entity_graph_db import EntityGraphDB

logger = setup_logger()


def extract_time_window(
    question: str,
    vlm: "VlModel",
    latest_frame: "Image | None" = None,
) -> float | None:
    """Extract time window from question using VLM with example-based learning.

    Uses a few example keywords as patterns, then asks VLM to extrapolate
    similar time references and return seconds.

    Args:
        question: User's question
        vlm: VLM instance to use for extraction
        latest_frame: Optional frame (required for VLM call, but image is ignored)

    Returns:
        Time window in seconds, or None if no time reference found
    """
    question_lower = question.lower()

    # Quick check for common patterns (fast path)
    if "last week" in question_lower or "past week" in question_lower:
        return 7 * 24 * 3600
    if "today" in question_lower or "last hour" in question_lower:
        return 3600
    if "recently" in question_lower or "recent" in question_lower:
        return 600

    # Use VLM to extract time reference from question
    # Provide examples and let VLM extrapolate similar patterns
    # Note: latest_frame is required by VLM interface but image content is ignored
    if not latest_frame:
        return None

    extraction_prompt = f"""Extract any time reference from this question and convert it to seconds.

Question: {question}

Examples of time references and their conversions:
- "last week" or "past week" -> 604800 seconds (7 days)
- "yesterday" -> 86400 seconds (1 day)
- "today" or "last hour" -> 3600 seconds (1 hour)
- "recently" or "recent" -> 600 seconds (10 minutes)
- "few minutes ago" -> 300 seconds (5 minutes)
- "just now" -> 60 seconds (1 minute)

Extrapolate similar patterns (e.g., "2 days ago", "this morning", "last month", etc.)
and convert to seconds. If no time reference is found, return "none".

Return ONLY a number (seconds) or the word "none". Do not include any explanation."""

    try:
        response = vlm.query(latest_frame, extraction_prompt)
        response = response.strip().lower()

        if "none" in response or not response:
            return None

        # Extract number from response
        numbers = re.findall(r"\d+(?:\.\d+)?", response)
        if numbers:
            seconds = float(numbers[0])
            # Sanity check: reasonable time windows (1 second to 1 year)
            if 1 <= seconds <= 365 * 24 * 3600:
                return seconds
    except Exception as e:
        logger.debug(f"Time extraction failed: {e}")

    return None


def build_graph_context(
    graph_db: "EntityGraphDB",
    entity_ids: list[str],
    time_window_s: float | None = None,
    max_relations_per_entity: int = 10,
    nearby_distance_meters: float = 5.0,
    current_video_time_s: float | None = None,
) -> dict[str, Any]:
    """Build enriched context from graph database for given entities.

    Args:
        graph_db: Entity graph database instance
        entity_ids: List of entity IDs to get context for
        time_window_s: Optional time window in seconds (e.g., 3600 for last hour)
        max_relations_per_entity: Maximum relations to include per entity (default: 10)
        nearby_distance_meters: Distance threshold for "nearby" entities (default: 5.0)
        current_video_time_s: Current video timestamp in seconds (for time window queries).
            If None, uses latest entity timestamp from DB as reference.

    Returns:
        Dictionary with graph context including relationships, distances, and semantics
    """
    if not graph_db or not entity_ids:
        return {}

    try:
        graph_context: dict[str, Any] = {
            "relationships": [],
            "spatial_info": [],
            "semantic_knowledge": [],
        }

        # Convert time_window_s to a (start_ts, end_ts) tuple if provided
        # Use video-relative timestamps, not wall-clock time
        time_window_tuple = None
        if time_window_s is not None:
            if current_video_time_s is not None:
                ref_time = current_video_time_s
            else:
                # Fallback: get the latest timestamp from entities in DB
                all_entities = graph_db.get_all_entities()
                ref_time = max((e.get("last_seen_ts", 0) for e in all_entities), default=0)
            time_window_tuple = (max(0, ref_time - time_window_s), ref_time)

        # Get recent relationships for each entity
        for entity_id in entity_ids:
            # Get relationships (Graph 1: interactions)
            relations = graph_db.get_relations_for_entity(
                entity_id=entity_id,
                relation_type=None,  # all types
                time_window=time_window_tuple,
            )
            for rel in relations[-max_relations_per_entity:]:
                graph_context["relationships"].append(
                    {
                        "subject": rel["subject_id"],
                        "relation": rel["relation_type"],
                        "object": rel["object_id"],
                        "confidence": rel["confidence"],
                        "when": rel["timestamp_s"],
                    }
                )

            # Get spatial relationships (Graph 2: distances)
            nearby = graph_db.get_nearby_entities(
                entity_id=entity_id, max_distance=nearby_distance_meters, latest_only=True
            )
            for dist in nearby:
                graph_context["spatial_info"].append(
                    {
                        "entity_a": entity_id,
                        "entity_b": dist["entity_id"],
                        "distance": dist.get("distance_meters"),
                        "category": dist.get("distance_category"),
                        "confidence": dist["confidence"],
                    }
                )

            # Get semantic knowledge (Graph 3: conceptual relations)
            semantic_rels = graph_db.get_semantic_relations(
                entity_id=entity_id,
                relation_type=None,
            )
            for sem in semantic_rels:
                graph_context["semantic_knowledge"].append(
                    {
                        "entity_a": sem["entity_a_id"],
                        "relation": sem["relation_type"],
                        "entity_b": sem["entity_b_id"],
                        "confidence": sem["confidence"],
                        "observations": sem["observation_count"],
                    }
                )

        # Get graph statistics for context
        if entity_ids:
            stats = graph_db.get_stats()
            graph_context["total_entities"] = stats.get("entities", 0)
            graph_context["total_relations"] = stats.get("relations", 0)

        return graph_context

    except Exception as e:
        logger.warning(f"failed to build graph context: {e}")
        return {}



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bool message type."""

from dimos_lcm.std_msgs import Bool as LCMBool

try:
    from std_msgs.msg import Bool as ROSBool  # type: ignore[attr-defined]
except ImportError:
    ROSBool = None  # type: ignore[assignment, misc]


class Bool(LCMBool):  # type: ignore[misc]
    """ROS-compatible Bool message."""

    msg_name = "std_msgs.Bool"

    def __init__(self, data: bool = False) -> None:
        """Initialize Bool with data value."""
        self.data = data

    @classmethod
    def from_ros_msg(cls, ros_msg: ROSBool) -> "Bool":
        """Create a Bool from a ROS std_msgs/Bool message.

        Args:
            ros_msg: ROS Bool message

        Returns:
            Bool instance
        """
        return cls(data=ros_msg.data)

    def to_ros_msg(self) -> ROSBool:
        """Convert to a ROS std_msgs/Bool message.

        Returns:
            ROS Bool message
        """
        if ROSBool is None:
            raise ImportError("ROS std_msgs not available")
        ros_msg = ROSBool()  # type: ignore[no-untyped-call]
        ros_msg.data = bool(self.data)
        return ros_msg



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Protocol

from dimos.core import Out
from dimos.msgs.sensor_msgs import CameraInfo, Image as ImageMsg, PointCloud2


class Image(Protocol):
    color_image: Out[ImageMsg]


class Camera(Image):
    camera_info: Out[CameraInfo]
    _camera_info: CameraInfo


class DepthCamera(Camera):
    depth_image: Out[ImageMsg]
    depth_camera_info: Out[CameraInfo]


class Pointcloud(Protocol):
    pointcloud: Out[PointCloud2]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Spatial vector database for storing and querying images with XY locations.

This module extends the ChromaDB implementation to support storing images with
their XY locations and querying by location or image similarity.
"""

from typing import Any

import chromadb
import numpy as np

from dimos.agents_deprecated.memory.visual_memory import VisualMemory
from dimos.types.robot_location import RobotLocation
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


class SpatialVectorDB:
    """
    A vector database for storing and querying images mapped to X,Y,theta absolute locations for SpatialMemory.

    This class extends the ChromaDB implementation to support storing images with
    their absolute locations and querying by location, text, or image cosine semantic similarity.
    """

    def __init__(  # type: ignore[no-untyped-def]
        self,
        collection_name: str = "spatial_memory",
        chroma_client=None,
        visual_memory=None,
        embedding_provider=None,
    ) -> None:
        """
        Initialize the spatial vector database.

        Args:
            collection_name: Name of the vector database collection
            chroma_client: Optional ChromaDB client for persistence. If None, an in-memory client is used.
            visual_memory: Optional VisualMemory instance for storing images. If None, a new one is created.
            embedding_provider: Optional ImageEmbeddingProvider instance for computing embeddings. If None, one will be created.
        """
        self.collection_name = collection_name

        # Use provided client or create in-memory client
        self.client = chroma_client if chroma_client is not None else chromadb.Client()

        # Check if collection already exists - in newer ChromaDB versions list_collections returns names directly
        existing_collections = self.client.list_collections()

        # Handle different versions of ChromaDB API
        try:
            collection_exists = collection_name in existing_collections
        except:
            try:
                collection_exists = collection_name in [c.name for c in existing_collections]
            except:
                try:
                    self.client.get_collection(name=collection_name)
                    collection_exists = True
                except Exception:
                    collection_exists = False

        # Get or create the collection
        self.image_collection = self.client.get_or_create_collection(
            name=collection_name, metadata={"hnsw:space": "cosine"}
        )

        # Use provided visual memory or create a new one
        self.visual_memory = visual_memory if visual_memory is not None else VisualMemory()

        # Store the embedding provider to reuse for all operations
        self.embedding_provider = embedding_provider

        # Initialize the location collection for text-based location tagging
        location_collection_name = f"{collection_name}_locations"
        self.location_collection = self.client.get_or_create_collection(
            name=location_collection_name, metadata={"hnsw:space": "cosine"}
        )

        # Log initialization info with details about whether using existing collection
        client_type = "persistent" if chroma_client is not None else "in-memory"
        try:
            count = len(self.image_collection.get(include=[])["ids"])
            if collection_exists:
                logger.info(
                    f"Using EXISTING {client_type} collection '{collection_name}' with {count} entries"
                )
            else:
                logger.info(f"Created NEW {client_type} collection '{collection_name}'")
        except Exception as e:
            logger.info(
                f"Initialized {client_type} collection '{collection_name}' (count error: {e!s})"
            )

    def add_image_vector(
        self,
        vector_id: str,
        image: np.ndarray,  # type: ignore[type-arg]
        embedding: np.ndarray,  # type: ignore[type-arg]
        metadata: dict[str, Any],
    ) -> None:
        """
        Add an image with its embedding and metadata to the vector database.

        Args:
            vector_id: Unique identifier for the vector
            image: The image to store
            embedding: The pre-computed embedding vector for the image
            metadata: Metadata for the image, including x, y coordinates
        """
        # Store the image in visual memory
        self.visual_memory.add(vector_id, image)

        # Add the vector to ChromaDB
        self.image_collection.add(
            ids=[vector_id], embeddings=[embedding.tolist()], metadatas=[metadata]
        )

        logger.info(f"Added image vector {vector_id} with metadata: {metadata}")

    def query_by_embedding(self, embedding: np.ndarray, limit: int = 5) -> list[dict]:  # type: ignore[type-arg]
        """
        Query the vector database for images similar to the provided embedding.

        Args:
            embedding: Query embedding vector
            limit: Maximum number of results to return

        Returns:
            List of results, each containing the image and its metadata
        """
        results = self.image_collection.query(
            query_embeddings=[embedding.tolist()], n_results=limit
        )

        return self._process_query_results(results)

    # TODO: implement efficient nearest neighbor search
    def query_by_location(
        self, x: float, y: float, radius: float = 2.0, limit: int = 5
    ) -> list[dict]:  # type: ignore[type-arg]
        """
        Query the vector database for images near the specified location.

        Args:
            x: X coordinate
            y: Y coordinate
            radius: Search radius in meters
            limit: Maximum number of results to return

        Returns:
            List of results, each containing the image and its metadata
        """
        results = self.image_collection.get()

        if not results or not results["ids"]:
            return []

        filtered_results = {"ids": [], "metadatas": [], "distances": []}  # type: ignore[var-annotated]

        for i, metadata in enumerate(results["metadatas"]):  # type: ignore[arg-type]
            item_x = metadata.get("x")
            item_y = metadata.get("y")

            if item_x is not None and item_y is not None:
                distance = np.sqrt((x - item_x) ** 2 + (y - item_y) ** 2)

                if distance <= radius:
                    filtered_results["ids"].append(results["ids"][i])
                    filtered_results["metadatas"].append(metadata)
                    filtered_results["distances"].append(distance)

        sorted_indices = np.argsort(filtered_results["distances"])
        filtered_results["ids"] = [filtered_results["ids"][i] for i in sorted_indices[:limit]]
        filtered_results["metadatas"] = [
            filtered_results["metadatas"][i] for i in sorted_indices[:limit]
        ]
        filtered_results["distances"] = [
            filtered_results["distances"][i] for i in sorted_indices[:limit]
        ]

        return self._process_query_results(filtered_results)

    def _process_query_results(self, results) -> list[dict]:  # type: ignore[no-untyped-def, type-arg]
        """Process query results to include decoded images."""
        if not results or not results["ids"]:
            return []

        processed_results = []

        for i, vector_id in enumerate(results["ids"]):
            if isinstance(vector_id, list) and not vector_id:
                continue

            lookup_id = vector_id[0] if isinstance(vector_id, list) else vector_id

            # Create the result dictionary with metadata regardless of image availability
            result = {
                "metadata": results["metadatas"][i] if "metadatas" in results else {},
                "id": lookup_id,
            }

            # Add distance if available
            if "distances" in results:
                result["distance"] = (
                    results["distances"][i][0]
                    if isinstance(results["distances"][i], list)
                    else results["distances"][i]
                )

            # Get the image from visual memory
            #image = self.visual_memory.get(lookup_id)
            #result["image"] = image

            processed_results.append(result)

        return processed_results

    def query_by_text(self, text: str, limit: int = 5) -> list[dict]:  # type: ignore[type-arg]
        """
        Query the vector database for images matching the provided text description.

        This method uses CLIP's text-to-image matching capability to find images
        that semantically match the text query (e.g., "where is the kitchen").

        Args:
            text: Text query to search for
            limit: Maximum number of results to return

        Returns:
            List of results, each containing the image, its metadata, and similarity score
        """
        if self.embedding_provider is None:
            from dimos.agents_deprecated.memory.image_embedding import ImageEmbeddingProvider

            self.embedding_provider = ImageEmbeddingProvider(model_name="clip")

        text_embedding = self.embedding_provider.get_text_embedding(text)

        results = self.image_collection.query(
            query_embeddings=[text_embedding.tolist()],
            n_results=limit,
            include=["documents", "metadatas", "distances"],
        )

        logger.info(
            f"Text query: '{text}' returned {len(results['ids'] if 'ids' in results else [])} results"
        )
        return self._process_query_results(results)

    def get_all_locations(self) -> list[tuple[float, float, float]]:
        """Get all locations stored in the database."""
        # Get all items from the collection without embeddings
        results = self.image_collection.get(include=["metadatas"])

        if not results or "metadatas" not in results or not results["metadatas"]:
            return []

        # Extract x, y coordinates from metadata
        locations = []
        for metadata in results["metadatas"]:
            if isinstance(metadata, list) and metadata and isinstance(metadata[0], dict):
                metadata = metadata[0]  # Handle nested metadata

            if isinstance(metadata, dict) and "x" in metadata and "y" in metadata:
                x = metadata.get("x", 0)
                y = metadata.get("y", 0)
                z = metadata.get("z", 0) if "z" in metadata else 0
                locations.append((x, y, z))

        return locations

    @property
    def image_storage(self):  # type: ignore[no-untyped-def]
        """Legacy accessor for compatibility with existing code."""
        return self.visual_memory.images

    def tag_location(self, location: RobotLocation) -> None:
        """
        Tag a location with a semantic name/description for text-based retrieval.

        Args:
            location: RobotLocation object with position/rotation data
        """

        location_id = location.location_id
        metadata = location.to_vector_metadata()

        self.location_collection.add(
            ids=[location_id], documents=[location.name], metadatas=[metadata]
        )

    def query_tagged_location(self, query: str) -> tuple[RobotLocation | None, float]:
        """
        Query for a tagged location using semantic text search.

        Args:
            query: Natural language query (e.g., "dining area", "place to eat")

        Returns:
            The best matching RobotLocation or None if no matches found
        """

        results = self.location_collection.query(
            query_texts=[query], n_results=1, include=["metadatas", "documents", "distances"]
        )

        if not (results and results["ids"] and len(results["ids"][0]) > 0):
            return None, 0

        best_match_metadata = results["metadatas"][0][0]  # type: ignore[index]
        distance = float(results["distances"][0][0] if "distances" in results else 0.0)  # type: ignore[index]

        location = RobotLocation.from_vector_metadata(best_match_metadata)  # type: ignore[arg-type]

        logger.info(
            f"Found location '{location.name}' for query '{query}' (distance: {distance:.3f})"
            if distance
            else ""
        )

        return location, distance



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

from dimos.core.o3dpickle import register_picklers

if TYPE_CHECKING:
    from collections.abc import Callable

# injects pickling system into o3d
register_picklers()
T = TypeVar("T")


def rpc(fn: Callable[..., Any]) -> Callable[..., Any]:
    fn.__rpc__ = True  # type: ignore[attr-defined]
    return fn



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable
from copy import copy
import threading
import time
from typing import Any

from dimos_lcm.foxglove_msgs.ImageAnnotations import (
    ImageAnnotations,
)
from lcm_msgs.foxglove_msgs import SceneUpdate  # type: ignore[import-not-found]
from reactivex.observable import Observable

from dimos.core import In, Out, rpc
from dimos.msgs.geometry_msgs import PoseStamped, Quaternion, Transform, Vector3
from dimos.msgs.sensor_msgs import Image, PointCloud2
from dimos.msgs.vision_msgs import Detection2DArray
from dimos.perception.detection.module3D import Detection3DModule
from dimos.perception.detection.type import ImageDetections3DPC, TableStr
from dimos.perception.detection.type.detection3d import Detection3DPC


# Represents an object in space, as collection of 3d detections over time
class Object3D(Detection3DPC):
    best_detection: Detection3DPC | None = None
    center: Vector3 | None = None  # type: ignore[assignment]
    track_id: str | None = None  # type: ignore[assignment]
    detections: int = 0

    def to_repr_dict(self) -> dict[str, Any]:
        if self.center is None:
            center_str = "None"
        else:
            center_str = (
                "[" + ", ".join(list(map(lambda n: f"{n:1f}", self.center.to_list()))) + "]"
            )
        return {
            "object_id": self.track_id,
            "detections": self.detections,
            "center": center_str,
        }

    def __init__(  # type: ignore[no-untyped-def]
        self, track_id: str, detection: Detection3DPC | None = None, *args, **kwargs
    ) -> None:
        if detection is None:
            return
        self.ts = detection.ts
        self.track_id = track_id
        self.class_id = detection.class_id
        self.name = detection.name
        self.confidence = detection.confidence
        self.pointcloud = detection.pointcloud
        self.bbox = detection.bbox
        self.transform = detection.transform
        self.center = detection.center
        self.frame_id = detection.frame_id
        self.detections = self.detections + 1
        self.best_detection = detection

    def __add__(self, detection: Detection3DPC) -> "Object3D":
        if self.track_id is None:
            raise ValueError("Cannot add detection to object with None track_id")
        new_object = Object3D(self.track_id)
        new_object.bbox = detection.bbox
        new_object.confidence = max(self.confidence, detection.confidence)
        new_object.ts = max(self.ts, detection.ts)
        new_object.track_id = self.track_id
        new_object.class_id = self.class_id
        new_object.name = self.name
        new_object.transform = self.transform
        new_object.pointcloud = self.pointcloud + detection.pointcloud
        new_object.frame_id = self.frame_id
        new_object.center = (self.center + detection.center) / 2
        new_object.detections = self.detections + 1

        if detection.bbox_2d_volume() > self.bbox_2d_volume():
            new_object.best_detection = detection
        else:
            new_object.best_detection = self.best_detection

        return new_object

    def get_image(self) -> Image | None:
        return self.best_detection.image if self.best_detection else None

    def scene_entity_label(self) -> str:
        return f"{self.name} ({self.detections})"

    def agent_encode(self):  # type: ignore[no-untyped-def]
        return {
            "id": self.track_id,
            "name": self.name,
            "detections": self.detections,
            "last_seen": f"{round(time.time() - self.ts)}s ago",
            # "position": self.to_pose().position.agent_encode(),
        }

    def to_pose(self) -> PoseStamped:
        if self.best_detection is None or self.center is None:
            raise ValueError("Cannot compute pose without best_detection and center")

        optical_inverse = Transform(
            translation=Vector3(0.0, 0.0, 0.0),
            rotation=Quaternion(-0.5, 0.5, -0.5, 0.5),
            frame_id="camera_link",
            child_frame_id="camera_optical",
        ).inverse()

        print("transform is", self.best_detection.transform)

        global_transform = optical_inverse + self.best_detection.transform

        print("inverse optical is", global_transform)

        print("obj center is", self.center)
        global_pose = global_transform.to_pose()
        print("Global pose:", global_pose)
        global_pose.frame_id = self.best_detection.frame_id
        print("remap to", self.best_detection.frame_id)
        return PoseStamped(
            position=self.center, orientation=Quaternion(), frame_id=self.best_detection.frame_id
        )


class ObjectDBModule(Detection3DModule, TableStr):
    cnt: int = 0
    objects: dict[str, Object3D]
    object_stream: Observable[Object3D] | None = None

    goto: Callable[[PoseStamped], Any] | None = None

    color_image: In[Image]
    pointcloud: In[PointCloud2]

    detections: Out[Detection2DArray]
    annotations: Out[ImageAnnotations]

    detected_pointcloud_0: Out[PointCloud2]
    detected_pointcloud_1: Out[PointCloud2]
    detected_pointcloud_2: Out[PointCloud2]

    detected_image_0: Out[Image]
    detected_image_1: Out[Image]
    detected_image_2: Out[Image]

    scene_update: Out[SceneUpdate]

    target: Out[PoseStamped]

    remembered_locations: dict[str, PoseStamped]

    @rpc
    def start(self) -> None:
        Detection3DModule.start(self)

        def update_objects(imageDetections: ImageDetections3DPC) -> None:
            for detection in imageDetections.detections:
                self.add_detection(detection)

        def scene_thread() -> None:
            while True:
                scene_update = self.to_foxglove_scene_update()
                self.scene_update.publish(scene_update)
                time.sleep(1.0)

        threading.Thread(target=scene_thread, daemon=True).start()

        self.detection_stream_3d.subscribe(update_objects)

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)
        self.goto = None
        self.objects = {}
        self.remembered_locations = {}

    def closest_object(self, detection: Detection3DPC) -> Object3D | None:
        # Filter objects to only those with matching names
        matching_objects = [obj for obj in self.objects.values() if obj.name == detection.name]

        if not matching_objects:
            return None

        # Sort by distance
        distances = sorted(matching_objects, key=lambda obj: detection.center.distance(obj.center))

        return distances[0]

    def add_detections(self, detections: list[Detection3DPC]) -> list[Object3D]:
        return [
            detection for detection in map(self.add_detection, detections) if detection is not None
        ]

    def add_detection(self, detection: Detection3DPC):  # type: ignore[no-untyped-def]
        """Add detection to existing object or create new one."""
        closest = self.closest_object(detection)
        if closest and closest.bounding_box_intersects(detection):
            return self.add_to_object(closest, detection)
        else:
            return self.create_new_object(detection)

    def add_to_object(self, closest: Object3D, detection: Detection3DPC):  # type: ignore[no-untyped-def]
        new_object = closest + detection
        if closest.track_id is not None:
            self.objects[closest.track_id] = new_object
        return new_object

    def create_new_object(self, detection: Detection3DPC):  # type: ignore[no-untyped-def]
        new_object = Object3D(f"obj_{self.cnt}", detection)
        if new_object.track_id is not None:
            self.objects[new_object.track_id] = new_object
        self.cnt += 1
        return new_object

    def agent_encode(self) -> str:
        ret = []
        for obj in copy(self.objects).values():
            # we need at least 3 detectieons to consider it a valid object
            # for this to be serious we need a ratio of detections within the window of observations
            if len(obj.detections) < 4:  # type: ignore[arg-type]
                continue
            ret.append(str(obj.agent_encode()))  # type: ignore[no-untyped-call]
        if not ret:
            return "No objects detected yet."
        return "\n".join(ret)

    # @rpc
    # def vlm_query(self, description: str) -> Object3D | None:
    #     imageDetections2D = super().ask_vlm(description)
    #     print("VLM query found", imageDetections2D, "detections")
    #     time.sleep(3)

    #     if not imageDetections2D.detections:
    #         return None

    #     ret = []
    #     for obj in self.objects.values():
    #         if obj.ts != imageDetections2D.ts:
    #             print(
    #                 "Skipping",
    #                 obj.track_id,
    #                 "ts",
    #                 obj.ts,
    #                 "!=",
    #                 imageDetections2D.ts,
    #             )
    #             continue
    #         if obj.class_id != -100:
    #             continue
    #         if obj.name != imageDetections2D.detections[0].name:
    #             print("Skipping", obj.name, "!=", imageDetections2D.detections[0].name)
    #             continue
    #         ret.append(obj)
    #     ret.sort(key=lambda x: x.ts)

    #     return ret[0] if ret else None

    def lookup(self, label: str) -> list[Detection3DPC]:
        """Look up a detection by label."""
        return []

    @rpc
    def stop(self):  # type: ignore[no-untyped-def]
        return super().stop()

    def goto_object(self, object_id: str) -> Object3D | None:
        """Go to object by id."""
        return self.objects.get(object_id, None)

    def to_foxglove_scene_update(self) -> "SceneUpdate":
        """Convert all detections to a Foxglove SceneUpdate message.

        Returns:
            SceneUpdate containing SceneEntity objects for all detections
        """

        # Create SceneUpdate message with all detections
        scene_update = SceneUpdate()
        scene_update.deletions_length = 0
        scene_update.deletions = []
        scene_update.entities = []

        for obj in self.objects:
            try:
                scene_update.entities.append(
                    obj.to_foxglove_scene_entity(entity_id=f"{obj.name}_{obj.track_id}")  # type: ignore[attr-defined]
                )
            except Exception:
                pass

        scene_update.entities_length = len(scene_update.entities)
        return scene_update

    def __len__(self) -> int:
        return len(self.objects.values())


detectionDB_module = ObjectDBModule.blueprint

__all__ = ["ObjectDBModule", "detectionDB_module"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import difflib
import math
import time

from unitree_webrtc_connect.constants import RTC_TOPIC

from dimos.core.core import rpc
from dimos.core.skill_module import SkillModule
from dimos.msgs.geometry_msgs import PoseStamped, Quaternion, Vector3
from dimos.navigation.base import NavigationState
from dimos.protocol.skill.skill import skill
from dimos.protocol.skill.type import Reducer, Stream
from dimos.robot.unitree_webrtc.unitree_skills import UNITREE_WEBRTC_CONTROLS
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


_UNITREE_COMMANDS = {
    name: (id_, description)
    for name, id_, description in UNITREE_WEBRTC_CONTROLS
    if name not in ["Reverse", "Spin"]
}


class UnitreeSkillContainer(SkillModule):
    """Container for Unitree Go2 robot skills using the new framework."""

    rpc_calls: list[str] = [
        "NavigationInterface.set_goal",
        "NavigationInterface.get_state",
        "NavigationInterface.is_goal_reached",
        "NavigationInterface.cancel_goal",
        "GO2Connection.publish_request",
    ]

    @rpc
    def start(self) -> None:
        super().start()
        # Initialize TF early so it can start receiving transforms.
        _ = self.tf

    @rpc
    def stop(self) -> None:
        super().stop()

    @skill()
    def relative_move(self, forward: float = 0.0, left: float = 0.0, degrees: float = 0.0) -> str:
        """Move the robot relative to its current position.

        The `degrees` arguments refers to the rotation the robot should be at the end, relative to its current rotation.

        Example calls:

            # Move to a point that's 2 meters forward and 1 to the right.
            relative_move(forward=2, left=-1, degrees=0)

            # Move back 1 meter, while still facing the same direction.
            relative_move(forward=-1, left=0, degrees=0)

            # Rotate 90 degrees to the right (in place)
            relative_move(forward=0, left=0, degrees=-90)

            # Move 3 meters left, and face that direction
            relative_move(forward=0, left=3, degrees=90)
        """
        forward, left, degrees = float(forward), float(left), float(degrees)

        tf = self.tf.get("world", "base_link")
        if tf is None:
            return "Failed to get the position of the robot."

        try:
            set_goal_rpc, get_state_rpc, is_goal_reached_rpc = self.get_rpc_calls(
                "NavigationInterface.set_goal",
                "NavigationInterface.get_state",
                "NavigationInterface.is_goal_reached",
            )
        except Exception:
            logger.error("Navigation module not connected properly")
            return "Failed to connect to navigation module."

        # TODO: Improve this. This is not a nice way to do it. I should
        # subscribe to arrival/cancellation events instead.

        set_goal_rpc(self._generate_new_goal(tf.to_pose(), forward, left, degrees))

        time.sleep(1.0)

        start_time = time.monotonic()
        timeout = 100.0
        while get_state_rpc() == NavigationState.FOLLOWING_PATH:
            if time.monotonic() - start_time > timeout:
                return "Navigation timed out"
            time.sleep(0.1)

        time.sleep(1.0)

        if not is_goal_reached_rpc():
            return "Navigation was cancelled or failed"
        else:
            return "Navigation goal reached"

    def _generate_new_goal(
        self, current_pose: PoseStamped, forward: float, left: float, degrees: float
    ) -> PoseStamped:
        local_offset = Vector3(forward, left, 0)
        global_offset = current_pose.orientation.rotate_vector(local_offset)
        goal_position = current_pose.position + global_offset

        current_euler = current_pose.orientation.to_euler()
        goal_yaw = current_euler.yaw + math.radians(degrees)
        goal_euler = Vector3(current_euler.roll, current_euler.pitch, goal_yaw)
        goal_orientation = Quaternion.from_euler(goal_euler)

        return PoseStamped(position=goal_position, orientation=goal_orientation)

    @skill()
    def wait(self, seconds: float) -> str:
        """Wait for a specified amount of time.

        Args:
            seconds: Seconds to wait
        """
        time.sleep(seconds)
        return f"Wait completed with length={seconds}s"

    @skill(stream=Stream.passive, reducer=Reducer.latest, hide_skill=True)  # type: ignore[arg-type]
    def current_time(self):  # type: ignore[no-untyped-def]
        """Provides current time implicitly, don't call this skill directly."""
        print("Starting current_time skill")
        while True:
            yield str(datetime.datetime.now())
            time.sleep(1)

    @skill()
    def execute_sport_command(self, command_name: str) -> str:
        try:
            publish_request = self.get_rpc_calls("GO2Connection.publish_request")
        except Exception:
            logger.error("GO2Connection not connected properly")
            return "Failed to connect to GO2Connection."

        if command_name not in _UNITREE_COMMANDS:
            suggestions = difflib.get_close_matches(
                command_name, _UNITREE_COMMANDS.keys(), n=3, cutoff=0.6
            )
            return f"There's no '{command_name}' command. Did you mean: {suggestions}"

        id_, _ = _UNITREE_COMMANDS[command_name]

        try:
            publish_request(RTC_TOPIC["SPORT_MOD"], {"api_id": id_})
            return f"'{command_name}' command executed successfully."
        except Exception as e:
            logger.error(f"Failed to execute {command_name}: {e}")
            return "Failed to execute the command."


_commands = "\n".join(
    [f'- "{name}": {description}' for name, (_, description) in _UNITREE_COMMANDS.items()]
)

UnitreeSkillContainer.execute_sport_command.__doc__ = f"""Execute a Unitree sport command.

Example usage:

    execute_sport_command("FrontPounce")

Here are all the command names and what they do.

{_commands}
"""


unitree_skills = UnitreeSkillContainer.blueprint

__all__ = ["UnitreeSkillContainer", "unitree_skills"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Event, RLock
import time
from typing import TYPE_CHECKING

import numpy as np
from reactivex.disposable import Disposable

from dimos.core.core import rpc
from dimos.core.global_config import GlobalConfig
from dimos.core.skill_module import SkillModule
from dimos.core.stream import In, Out
from dimos.models.qwen.video_query import BBox
from dimos.models.segmentation.edge_tam import EdgeTAMProcessor
from dimos.models.vl.qwen import QwenVlModel
from dimos.msgs.geometry_msgs import Twist
from dimos.msgs.sensor_msgs import CameraInfo, Image, PointCloud2
from dimos.navigation.visual.query import get_object_bbox_from_image
from dimos.navigation.visual_servoing.detection_navigation import DetectionNavigation
from dimos.navigation.visual_servoing.visual_servoing_2d import VisualServoing2D
from dimos.protocol.skill.skill import skill
from dimos.utils.logging_config import setup_logger

if TYPE_CHECKING:
    from dimos.models.vl.base import VlModel

logger = setup_logger()


class PersonFollowSkillContainer(SkillModule):
    """Skill container for following a person.

    This skill uses:
    - A VL model (QwenVlModel) to initially detect a person from a text description.
    - EdgeTAM for continuous tracking across frames.
    - Visual servoing OR 3D navigation to control robot movement towards the person.
    - Does not do obstacle avoidance; assumes a clear path.
    """

    color_image: In[Image]
    global_map: In[PointCloud2]
    cmd_vel: Out[Twist]

    _frequency: float = 20.0  # Hz - control loop frequency
    _max_lost_frames: int = 15  # number of frames to wait before declaring person lost

    def __init__(
        self,
        camera_info: CameraInfo,
        global_config: GlobalConfig,
        use_3d_navigation: bool = False,
    ) -> None:
        super().__init__()
        self._global_config: GlobalConfig = global_config
        self._use_3d_navigation: bool = use_3d_navigation
        self._latest_image: Image | None = None
        self._latest_pointcloud: PointCloud2 | None = None
        self._vl_model: VlModel = QwenVlModel()
        self._tracker: EdgeTAMProcessor | None = None
        self._should_stop: Event = Event()
        self._lock = RLock()

        # Use MuJoCo camera intrinsics in simulation mode
        if self._global_config.simulation:
            from dimos.robot.unitree_webrtc.mujoco_connection import MujocoConnection

            camera_info = MujocoConnection.camera_info_static

        self._camera_info = camera_info
        self._visual_servo = VisualServoing2D(camera_info, self._global_config.simulation)
        self._detection_navigation = DetectionNavigation(self.tf, camera_info)

    @rpc
    def start(self) -> None:
        super().start()
        self._disposables.add(Disposable(self.color_image.subscribe(self._on_color_image)))
        if self._use_3d_navigation:
            self._disposables.add(Disposable(self.global_map.subscribe(self._on_pointcloud)))

    @rpc
    def stop(self) -> None:
        self._stop_following()

        with self._lock:
            if self._tracker is not None:
                self._tracker.stop()
                self._tracker = None

        self._vl_model.stop()
        super().stop()

    @skill()
    def follow_person(self, query: str) -> str:
        """Follow a person matching the given description using visual servoing.

        The robot will continuously track and follow the person, while keeping
        them centered in the camera view.

        Args:
            query: Description of the person to follow (e.g., "man with blue shirt")

        Returns:
            Status message indicating the result of the following action.

        Example:
            follow_person("man with blue shirt")
            follow_person("person in the doorway")
        """

        self._stop_following()

        self._should_stop.clear()

        with self._lock:
            latest_image = self._latest_image

        if latest_image is None:
            return "No image available to detect person."

        initial_bbox = get_object_bbox_from_image(
            self._vl_model,
            latest_image,
            query,
        )

        if initial_bbox is None:
            return f"Could not find '{query}' in the current view."

        return self._follow_loop(query, initial_bbox)

    @skill()
    def stop_following(self) -> str:
        """Stop following the current person.

        Returns:
            Confirmation message.
        """
        self._stop_following()

        self.cmd_vel.publish(Twist.zero())

        return "Stopped following."

    def _on_color_image(self, image: Image) -> None:
        with self._lock:
            self._latest_image = image

    def _on_pointcloud(self, pointcloud: PointCloud2) -> None:
        with self._lock:
            self._latest_pointcloud = pointcloud

    def _follow_loop(self, query: str, initial_bbox: BBox) -> str:
        x1, y1, x2, y2 = initial_bbox
        box = np.array([x1, y1, x2, y2], dtype=np.float32)

        with self._lock:
            if self._tracker is None:
                self._tracker = EdgeTAMProcessor()
            tracker = self._tracker
            latest_image = self._latest_image
            if latest_image is None:
                return "No image available to start tracking."

        initial_detections = tracker.init_track(
            image=latest_image,
            box=box,
            obj_id=1,
        )

        if len(initial_detections) == 0:
            self.cmd_vel.publish(Twist.zero())
            return f"EdgeTAM failed to segment '{query}'."

        logger.info(f"EdgeTAM initialized with {len(initial_detections)} detections")

        lost_count = 0
        period = 1.0 / self._frequency
        next_time = time.monotonic()

        while not self._should_stop.is_set():
            next_time += period

            with self._lock:
                latest_image = self._latest_image
                assert latest_image is not None

            detections = tracker.process_image(latest_image)

            if len(detections) == 0:
                self.cmd_vel.publish(Twist.zero())

                lost_count += 1
                if lost_count > self._max_lost_frames:
                    self.cmd_vel.publish(Twist.zero())
                    return f"Lost track of '{query}'. Stopping."
            else:
                lost_count = 0
                best_detection = max(detections.detections, key=lambda d: d.bbox_2d_volume())

                if self._use_3d_navigation:
                    with self._lock:
                        pointcloud = self._latest_pointcloud
                    if pointcloud is None:
                        self.cmd_vel.publish(Twist.zero())
                        return "No pointcloud available for 3D navigation. Stopping."
                    twist = self._detection_navigation.compute_twist_for_detection_3d(
                        pointcloud,
                        best_detection,
                        latest_image,
                    )
                    if twist is None:
                        self.cmd_vel.publish(Twist.zero())
                        return f"3D navigation failed for '{query}'. Stopping."
                else:
                    twist = self._visual_servo.compute_twist(
                        best_detection.bbox,
                        latest_image.width,
                    )
                self.cmd_vel.publish(twist)

            now = time.monotonic()
            sleep_duration = next_time - now
            if sleep_duration > 0:
                time.sleep(sleep_duration)

        self.cmd_vel.publish(Twist.zero())
        return "Stopped following as requested."

    def _stop_following(self) -> None:
        self._should_stop.set()


person_follow_skill = PersonFollowSkillContainer.blueprint

__all__ = ["PersonFollowSkillContainer", "person_follow_skill"]
